#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
import humanize
from pathlib import Path
from collections import defaultdict

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)

# Inspired by: https://github.com/tum-ei-eda/etiss/blob/master/src/bare_etiss_processor/get_metrics.py


# TODO: merge mem_trace with instr_trace
# TODO: trunc_trace also for mem_trace?


DEFAULT_STACK_SIZE = 0x4000


def helper(x):
    return int(x, 16)


def init_mems(romStart, ramStart, heapStart, ramSize, stackSize):
    r = MemRange("ROM", romStart, ramStart)
    d = MemRange("Data", ramStart, heapStart)
    h = MemRange("Heap", heapStart, ramStart + ramSize - stackSize)
    s = MemRange("Stack", ramStart + ramSize - stackSize, ramStart + ramSize)
    mems = [r, d, h, s]
    return mems


def worker(args):
    addrs, romStart, ramStart, heapStart, ramSize, stackSize = args
    mems = init_mems(romStart, ramStart, heapStart, ramSize, stackSize)
    for addr, mode, pc, sz in addrs:
        # addr = int(addr, 16)
        for mem in mems:
            if mem.contains(addr):
                mem.trace(addr, mode, pc, sz)
    mems_ = []
    for mem in mems:
        mems_.append(mem.freeze())
    return mems_


def print_sz(sz, unknown_msg=""):
    if sz is None:
        return f"unknown [{unknown_msg}]" if unknown_msg else "unknown"
    return humanize.naturalsize(sz) + " (" + hex(sz) + ")"


class MemRange:
    def __init__(self, name, min, max):
        self.name = name
        self.min = min
        self.max = max
        assert self.min <= self.max, "Invalid MemRange"
        self.num_reads = 0
        self.num_reads_per_size = defaultdict(int)
        self.read_alignments_per_size = defaultdict(lambda: defaultdict(int))
        self.num_writes = 0
        self.num_writes_per_size = defaultdict(int)
        self.write_alignments_per_size = defaultdict(lambda: defaultdict(int))
        self.read_bytes = 0
        self.read_bytes_per_size = defaultdict(int)
        self.read_bytes_alignments_per_size = defaultdict(lambda: defaultdict(int))
        self.written_bytes = 0
        self.written_bytes_per_size = defaultdict(int)
        self.written_bytes_alignments_per_size = defaultdict(lambda: defaultdict(int))
        self.low = 0xFFFFFFFF
        self.high = 0
        self.alignments = set()

    def contains(self, adr):
        return adr >= self.min and adr < self.max

    def update(self, other):
        self.low = min(other.low, self.low)
        self.high = max(other.high, self.high)
        self.num_reads += other.num_reads
        self.num_writes += other.num_writes
        self.read_bytes += other.read_bytes
        self.written_bytes += other.written_bytes
        self.alignments |= other.alignments
        for sz, num in other.num_reads_per_size.items():
            self.num_reads_per_size[sz] += num
        for sz, num in other.num_writes_per_size.items():
            self.num_writes_per_size[sz] += num
        for sz, num in other.read_bytes_per_size.items():
            self.read_bytes_per_size[sz] += num
        for sz, num in other.written_bytes_per_size.items():
            self.written_bytes_per_size[sz] += num

    def trace(self, adr, mode, pc, sz):
        self.low = min(adr, self.low)
        self.high = max(adr, self.high)
        alignment = adr % sz
        self.alignments.add(alignment)
        if mode == "r":
            self.num_reads += 1
            self.num_reads_per_size[sz] += 1
            self.read_alignments_per_size[sz][alignment] += 1
            self.read_bytes += sz
            self.read_bytes_per_size[sz] += sz
            self.read_bytes_alignments_per_size[sz][alignment] += sz
        elif mode == "w":
            self.num_writes += 1
            self.num_writes_per_size[sz] += 1
            self.write_alignments_per_size[sz][alignment] += 1
            self.written_bytes += sz
            self.written_bytes_per_size[sz] += sz
            self.written_bytes_alignments_per_size[sz][alignment] += sz
        else:
            raise ValueError(f"Invalid mode: {mode}")

    @property
    def sizes(self):
        sizes = sorted(list(set(self.num_reads_per_size.keys()) | set(self.num_writes_per_size.keys())))
        return sizes

    @property
    def count(self):
        return self.num_reads + self.num_writes

    @property
    def count_per_size(self):
        return {size: self.num_reads_per_size.get(size, 0) + self.num_writes_per_size(size, 0) for size in self.sizes}

    def usage(self):
        if self.low > self.high:
            return 0
        return self.high - self.low

    def stats(self):
        if self.low > self.high:
            return self.name + "\t[not accessed]"
        return (
            f"{self.name}\t[0x{self.low:x}-0x{self.high:x}] \t({self.count} times, "
            f"reads: {self.num_reads} <{self.read_bytes}B>, "
            f"writes: {self.num_writes} <{self.written_bytes}B>)"
        )

    def freeze(self):
        return FrozenMemRange(self)


class FrozenMemRange:
    def __init__(self, mem: MemRange):
        self.name = mem.name
        self.min = mem.min
        self.max = mem.max
        self.num_reads = mem.num_reads
        self.num_reads_per_size = dict(mem.num_reads_per_size)
        self.read_alignments_per_size = {k: dict(v) for k, v in mem.read_alignments_per_size.items()}
        self.num_writes = mem.num_writes
        self.num_writes_per_size = dict(mem.num_writes_per_size)
        self.write_alignments_per_size = {k: dict(v) for k, v in mem.write_alignments_per_size.items()}
        self.read_bytes = mem.read_bytes
        self.read_bytes_per_size = dict(mem.read_bytes_per_size)
        self.read_bytes_alignments_per_size = {k: dict(v) for k, v in mem.read_bytes_alignments_per_size.items()}
        self.written_bytes = mem.written_bytes
        self.written_bytes_per_size = dict(mem.written_bytes_per_size)
        self.written_bytes_alignments_per_size = {k: dict(v) for k, v in mem.written_bytes_alignments_per_size.items()}
        self.low = mem.low
        self.high = mem.high
        self.alignments = mem.alignments

    def usage(self):
        if self.low > self.high:
            return 0
        return self.high - self.low


def process_symbol_table(symbol_table_df):
    # TODO: do not iterate over all symbols?
    heap_start = None
    for sym in symbol_table_df.itertuples(index=False):
        if sym.name == "_heap_start":
            heap_start = sym.value
    return heap_start


def process_sections(mem_sections_df):
    m = {}
    m["rom_rodata"] = 0
    m["rom_code"] = 0
    m["rom_misc"] = 0
    m["ram_data"] = 0
    m["ram_zdata"] = 0

    ignoreSections = [
        "",
        ".stack",
        ".comment",
        ".riscv.attributes",
        ".strtab",
        ".shstrtab",
    ]

    for s in mem_sections_df.itertuples(index=False):
        if s.name.startswith(".text"):
            m["rom_code"] += s.data_size
        elif s.name.startswith(".srodata"):
            m["rom_rodata"] += s.data_size
        elif s.name.startswith(".sdata"):
            m["ram_data"] += s.data_size
        elif s.name == ".rodata":
            m["rom_rodata"] += s.data_size
        elif s.name == ".vectors" or s.name == ".init_array":
            m["rom_misc"] += s.data_size
        elif s.name == ".data":
            m["ram_data"] += s.data_size
        elif s.name == ".bss" or s.name == ".sbss" or s.name == ".shbss":
            m["ram_zdata"] += s.data_size
        elif s.name.startswith(".gcc_except"):
            pass
        elif s.name.startswith(".sdata2"):
            pass
        elif s.name.startswith(".debug_"):
            pass
        elif s.name in ignoreSections:
            pass
        else:
            print("warning: ignored: " + s.name + " / size: " + str(s.data_size))

    return m


def collect_mem_metrics(
    mem_trace_df,
    mem_sections_df,
    symbol_table_df,
    mem_layout_df,
    max_stack: int = DEFAULT_STACK_SIZE,
    verbose: bool = False,
):
    # print("mem_trace_df", len(mem_trace_df))
    # TODO: count per function?

    static_sizes = process_sections(mem_sections_df)
    heap_start = process_symbol_table(symbol_table_df)

    rom_start = mem_layout_df[mem_layout_df["segment"] == "rom"]["start"].iloc[0]
    ram_start = mem_layout_df[mem_layout_df["segment"] == "ram"]["start"].iloc[0]
    ram_size = mem_layout_df[mem_layout_df["segment"] == "ram"]["size"].iloc[0]
    # heap_start = mem_layout_df["heap_start"].iloc[0]
    stack_size = max_stack

    mems = init_mems(rom_start, ram_start, heap_start, ram_size, stack_size)
    _, _, h, s = mems

    addrs = mem_trace_df[["addr", "mode", "pc", "bytes"]].values
    import multiprocessing as mp
    import os

    USE_POOL = False

    if USE_POOL:
        num_threads = os.cpu_count()
        with mp.Pool(num_threads) as p:
            num_chunks = num_threads

            def iter_chunks(lst, n):
                """Yield successive n-sized chunks from lst."""
                for i in range(0, len(lst), n):
                    yield lst[i : i + n]

            n = len(addrs) // num_chunks
            chunked_addrs = iter_chunks(addrs, n)
            args = [(chunk, rom_start, ram_start, heap_start, ram_size, stack_size) for chunk in chunked_addrs]
            for mems_ in p.imap_unordered(worker, args):
                for i, mem_ in enumerate(mems_):
                    if mem_.usage() == 0:
                        continue
                    mem = mems[i]
                    mem.update(mem_)
                    # mem.trace(mem_.high)
                    # mem.trace(mem_.low)
                    # mem.count += mem_.count
    else:
        for row in mem_trace_df.itertuples(index=False):
            # print("row", row)
            # input(">")
            for mem in mems:
                # print("mem", mem)
                if mem.contains(row.addr):
                    # print("trace")
                    mem.trace(row.addr, row.mode, row.pc, row.bytes)
    if verbose:
        for mem in mems:
            print(mem.stats())

    rom_size = sum([static_sizes[k] for k in static_sizes if k.startswith("rom_")])
    ram_size = sum([static_sizes[k] for k in static_sizes if k.startswith("ram_")])

    trace_available = True

    results = {
        "rom": rom_size,
        "rom_rodata": static_sizes["rom_rodata"],
        "rom_code": static_sizes["rom_code"],
        "rom_misc": static_sizes["rom_misc"],
        "ram": (ram_size + s.usage() + h.usage()) if trace_available else ram_size,
        "ram_data": static_sizes["ram_data"],
        "ram_zdata": static_sizes["ram_zdata"],
        "ram_stack": s.usage() if trace_available else None,
        "ram_heap": h.usage() if trace_available else None,
    }

    if verbose:
        print("=== Results ===")
        print("ROM usage:        " + print_sz(results["rom"]))
        print("  read-only data: " + print_sz(results["rom_rodata"]))
        print("  code:           " + print_sz(results["rom_code"]))
        print("  other required: " + print_sz(results["rom_misc"]))
        print(
            "RAM usage:        "
            + print_sz(results["ram"])
            + ("" if trace_available else " [stack and heap usage not included]")
        )
        print("  data:           " + print_sz(results["ram_data"]))
        print("  zero-init data: " + print_sz(results["ram_zdata"]))
        print("  stack:          " + print_sz(results["ram_stack"], unknown_msg="missing trace file"))
        print("  heap:           " + print_sz(results["ram_heap"], unknown_msg="missing trace file"))

    mem_metrics_data = [results]
    mem_metrics_df = pd.DataFrame(mem_metrics_data)

    results2 = [
        {
            "name": r.name,
            "low": r.low,
            "high": r.high,
            "count": r.count,
            "num_reads": r.num_reads,
            "num_writes": r.num_writes,
            "read_bytes": r.read_bytes,
            "written_bytes": r.written_bytes,
        }
        for r in mems
    ]
    mem_access_df = pd.DataFrame(results2)

    results3 = []
    for r in mems:
        temp = {
            "name": r.name,
            "low": r.low,
            "high": r.high,
        }
        for size in r.sizes:
            for align in sorted(list(r.alignments)):
                temp_ = {
                    **temp,
                    "size": size,
                    "alignment": align,
                    "count": r.count,
                    # "num_reads": r.num_reads_per_size.get(size, 0),
                    "num_reads": r.read_alignments_per_size.get(size, {}).get(align),
                    # "num_writes": r.num_writes_per_size.get(size, 0),
                    "num_writes": r.write_alignments_per_size.get(size, {}).get(align),
                    # "read_bytes": r.read_bytes_per_size.get(size, 0),
                    "read_bytes": r.read_bytes_alignments_per_size.get(size, {}).get(align),
                    # "written_bytes": r.written_bytes_per_size.get(size, 0),
                    "written_bytes": r.written_bytes_alignments_per_size.get(size, {}).get(align),
                }
                results3.append(temp_)
    mem_access_df2 = pd.DataFrame(results3)

    return mem_metrics_df, mem_access_df, mem_access_df2


def analyze_mem_trace(
    sess: Session,
    force: bool = False,
    max_stack: int = DEFAULT_STACK_SIZE,
    verbose: bool = False,
):
    artifacts = sess.artifacts
    # print("artifacts", artifacts)

    # Memory Trace
    mem_trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_trace")
    assert len(mem_trace_artifacts) == 1
    mem_trace_artifact = mem_trace_artifacts[0]
    assert mem_trace_artifact.attrs.get("simulator") in ["etiss"]  # TODO: support spike?

    # Memory Sections
    mem_sections_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_sections"
    )
    assert len(mem_sections_artifacts) == 1
    mem_sections_artifact = mem_sections_artifacts[0]

    # Symbol Table
    symbol_table_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "symbol_table"
    )
    assert len(symbol_table_artifacts) == 1
    symbol_table_artifact = symbol_table_artifacts[0]

    # Memory Layout
    mem_layout_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TABLE and x.name == "mem_layout"
    )
    assert len(mem_layout_artifacts) == 1
    mem_layout_artifact = mem_layout_artifacts[0]

    mem_metrics_df, mem_access_df, mem_access_df2 = collect_mem_metrics(
        mem_trace_artifact.df,
        mem_sections_artifact.df,
        symbol_table_artifact.df,
        mem_layout_artifact.df,
        max_stack=max_stack,
        verbose=verbose,
    )

    attrs = {
        "mem_trace": mem_trace_artifact.name,
        "kind": "metrics",
        "by": __name__,
    }

    mem_metrics_artifact = TableArtifact("mem_metrics", mem_metrics_df, attrs=attrs)
    sess.add_artifact(mem_metrics_artifact, override=force)

    attrs2 = {
        "mem_trace": mem_trace_artifact.name,
        "kind": "table",
        "by": __name__,
    }

    mem_access_artifact = TableArtifact("mem_access", mem_access_df, attrs=attrs2)
    sess.add_artifact(mem_access_artifact, override=force)
    mem_access_artifact2 = TableArtifact("mem_access2", mem_access_df2, attrs=attrs2)
    sess.add_artifact(mem_access_artifact2, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    analyze_mem_trace(sess, force=args.force, max_stack=args.max_stack, verbose=args.verbose)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--verbose", "-v", action="store_true")
    parser.add_argument("--max-stack", type=int, default=DEFAULT_STACK_SIZE)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import pickle
from pathlib import Path

import networkx as nx


def main():
    assert len(sys.argv) == 3, "Unexpected number of arguments"
    in_file = Path(sys.argv[1])
    assert in_file.is_file()
    assert in_file.suffix == ".pkl"
    out_file = Path(sys.argv[2])
    fmt = out_file.suffix[1:]
    assert fmt in ["dot", "png", "pdf"]
    with open(in_file, "rb") as f:
        graph = pickle.load(f)
    graph = nx.nx_agraph.to_agraph(graph)
    if fmt == "dot":
        graph.write(out_file)
    else:
        # prog = "neato"
        prog = "dot"
        graph.draw(out_file, prog=prog)


if __name__ == "__main__":
    main()



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import yaml
import pickle
import argparse
import multiprocessing
from typing import Optional, Union, List
from pathlib import Path
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

import m2isar
import m2isar.metamodel.arch
from m2isar.metamodel import M2_METAMODEL_VERSION, M2Model

from m2isar.frontends.coredsl2_set.parser import parse_cdsl2_set
from m2isar.backends.coredsl2_set.writer import gen_cdsl_code
from m2isar.transforms.encode_instructions.encoder import encode_instructions

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()

import logging

logging.getLogger("behav_builder").setLevel(logging.WARNING)
logging.getLogger("arch_builder").setLevel(logging.WARNING)
logging.getLogger("set_parser").setLevel(logging.WARNING)
logging.getLogger("visit_importer").setLevel(logging.WARNING)


def parse_generated_set(
    set_file,
    skip_errors: bool = False,
    extra_includes=None,
    add_mnemonic_prefix: bool = False,
    set_name: Optional[str] = None,
    auto_encoding: bool = False,
):
    try:
        models = parse_cdsl2_set(set_file, extra_includes)
    except Exception as e:
        if skip_errors:
            logger.exception(e)
            return None, str(e)
        else:
            raise e
    instr_sets = list(models.values())
    set_names = list(models.keys())
    instr_sets = [
        instr_set
        for instr_set in instr_sets
        if len(instr_set.instructions) > 0 or len(instr_set.unencoded_instructions) > 0
    ]
    if len(instr_sets) == 1:
        instr_set = instr_sets[0]
        if set_name is None:
            set_name = set_names[0]
    else:
        instr_set = instr_sets[-1]
        if set_name is None:
            set_name = set_names[-1]
    ret = {}
    if auto_encoding:
        assert len(instr_set.instructions) == 0
        assert len(instr_set.unencoded_instructions) > 0
        for instr_def in instr_set.unencoded_instructions.values():
            name = instr_def.name
            if add_mnemonic_prefix:
                prefix = set_name.lower()
                name_lower = name.lower()
                mnemonic = f"{prefix}.{name_lower}"
                instr_def.mnemonic = mnemonic
            assert name not in ret, f"Duplicate instrustion name: {name}"
            ret[name] = instr_def
    else:
        assert len(instr_set.instructions) > 0
        assert len(instr_set.unencoded_instructions) == 0
        for instr_def in instr_set.instructions.values():
            name = instr_def.name
            if add_mnemonic_prefix:
                prefix = set_name.lower()
                name_lower = name.lower()
                mnemonic = f"{prefix}.{name_lower}"
                instr_def.mnemonic = mnemonic
            assert name not in ret, f"Duplicate instrustion name: {name}"
            ret[name] = instr_def
    # contributing_types.append(set_name_)
    # break  # TODO
    return ret, set_name, None


def get_cdsl_sets(ext: str, xlen: int = 32, compressed: bool = False):
    ext = ext.lower()
    cdsl_sets = {
        "i": [f"RV{xlen}I"] + ([f"RV{xlen}IC"] if compressed else []),
        "m": [f"RV{xlen}M"],
        "a": [f"RV{xlen}A"],
        "f": [f"RV{xlen}F"] + ([f"RV{xlen}FC"] if compressed else []),
        "d": [f"RV{xlen}D"] + ([f"RV{xlen}DC"] if compressed else []),
        "zifencei": ["Zifencei"],
        "zicsr": ["Zicsr"],
        "zve32x": ["Zve32x"],
        "zve32f": ["Zve32f"],
    }
    ret = cdsl_sets.get(ext, None)
    assert ret is not None, f"Lookup failed for ext: {ext}"
    return ret


def apply_etiss_overrides(sets: List[str], semihosting: bool = True):
    etiss_overrides = {
        "Zicsr": "tum_csr",
        "RV32A": "tum_rva",
        "RV64A": "tum_rva64",
        "RV64M": "tum_rvm",
    }
    ret = []
    new = []
    for orig in sets:
        replaced = etiss_overrides.get(orig, None)
        if replaced:
            new.append(replaced)
        else:
            ret.append(orig)
    if semihosting:
        new += ["tum_semihosting"]
    new += ["tum_ret"]
    return ret + new


def get_cdsl_includes(
    sets: List[str],
    base_dir: Union[str, Path] = "rv_base",
    tum_dir: Union[str, Path] = ".",
):
    # Reverse lookup required!
    cdsl_includes = {
        "RISCVBase.core_desc": ["RISCVBase"],
        # "RV32I.core_desc": ["RV32I", "Zicsr", "Zifencei", "RVNMode", "RVSMode", "RVDebug"],
        "RV64I.core_desc": ["RV64I"],
        "RVA.core_desc": ["RV32A", "RV64A"],
        "RVC.core_desc": [
            "RV32IC",
            "RV32FC",
            "RV32DC",
            "RV64IC",
            "RV64FC",
            "RV64DC",
            "RV128IC",
        ],
        "RVD.core_desc": ["RV32D", "RV64D"],
        "RVF.core_desc": ["RV32F", "RV64F"],
        "RVI.core_desc": [
            "RV32I",
            "RV64I",
            "Zicsr",
            "Zifencei",
            "RVNMode",
            "RVSMode",
            "RVDebug",
        ],
        "RVM.core_desc": ["RV32M", "RV64M"],
        "tum_rvm.core_desc": ["tum_rvm"],
        "tum_rva.core_desc": ["tum_rva", "tum_rva64"],
        "tum_mod.core_desc": ["tum_ret", "tum_csr", "tum_semihosting"],
        "RVV.core_desc": [
            "RV32V",
            "RV64V",
            "Zve32x",
            "Zve32f",
            "Zve64x",
            "Zve64f",
            "Zve64d",
        ],
    }
    ret = set()
    for set_name in sets:
        # print("set_name", set_name)
        matched = None
        for cdsl_file, cdsl_sets in cdsl_includes.items():
            if set_name in cdsl_sets:
                matched = cdsl_file
                break
        assert matched, f"Include could not be resolved for set: {set_name}"
        if "tum" in set_name.lower():
            base = Path(tum_dir)
        else:
            base = Path(base_dir)
        ret.add(str(base / matched))
    return ret


def get_includes_code(includes: List[Union[str, Path]], prefix: str = ""):
    ret = []
    for inc in includes:
        if isinstance(inc, Path):
            inc = str(inc)
        if ".core_desc" not in inc:
            inc = f"{inc}.core_desc"
        if inc[0] != "/":
            inc = f"{prefix}{inc}"
        new = f'import "{inc}"'
        ret.append(new)
    return "\n".join(ret)


# def generate_base_core
#     core_name: str = "IsaacCore",
#     xlen: int = 32,
#     ignore_etiss: bool = False,
#     semihosting: bool = True,
#     base_extensions: List[str] = ["i", "m", "a", "f", "d", "c", "zifencei"],
#     # etiss_overrides: List[str] = ["tum_csr", "tum_ret", "tum_rva", "tum_semihosting"],
# ):


def generate_etiss_core(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    gen_dir: Optional[Union[str, Path]] = None,
    index_files: Optional[List[Union[str, Path]]] = None,
    core_name: str = "IsaacCore",
    set_names: List[str] = ["XIsaac"],
    xlen: int = 32,
    ignore_etiss: bool = False,
    semihosting: bool = True,
    base_extensions: List[str] = ["i", "m", "a", "f", "d", "c", "zifencei", "zicsr"],
    auto_encoding: bool = True,
    split: bool = True,  # One set per new instr
    force: bool = False,
    base_dir: str = "rv_base",
    tum_dir: str = ".",
    skip_errors: bool = True,
    extra_includes: Optional[List[Union[str, Path]]] = None,
    add_mnemonic_prefix: bool = False,
    n_parallel: int = 1,
):
    logger.info("Generating ETISS core...")
    # artifacts = sess.artifacts
    # TODO: get combined_index.yml from artifacts!
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    gen_dir = Path(gen_dir) if gen_dir is not None else workdir / "gen"
    gen_dir.mkdir(exist_ok=True)
    core_out_model_file = gen_dir / f"{core_name}.m2isarmodel"
    core_out_cdsl_file = gen_dir / f"{core_name}.core_desc"
    if index_files is None:
        combined_index_file = workdir / "combined_index.yml"  # if index_file is None else Path(index_file)
        index_files = [combined_index_file]
    else:
        assert isinstance(index_files, list)
        index_files = list(map(Path, index_files))
        assert len(index_files) > 0
        assert len(set_names) == len(index_files)
    generated_sets = []
    sub2files = defaultdict(dict)
    sub2new_files = defaultdict(dict)
    set_instr2sub_name = {}

    for k, combined_index_file in enumerate(index_files):
        set_name = set_names[k]
        assert combined_index_file.is_file()
        with open(combined_index_file, "r") as f:
            index_data = yaml.safe_load(f)
        for i, sub_data in enumerate(index_data["candidates"]):
            sub_artifacts = sub_data["artifacts"]
            # sub_properties = sub_data["properties"]
            sub_name = f"C{i}"
            sub_file = Path(sub_artifacts["cdsl"])
            sub2files[sub_name]["cdsl"] = sub_file
            sub_encoded_file = Path(sub_artifacts["cdsl_encoded"]) if "cdsl_encoded" in sub_artifacts else None
            sub_properties = sub_data["properties"]
            instr_name = sub_properties["InstrName"]
            if sub_encoded_file is not None:
                sub2files[sub_name]["cdsl_encoded"] = sub_encoded_file
            set_instr2sub_name[(set_name, instr_name)] = sub_name
            generated_sets.append((set_name, sub_name, sub_file, sub_encoded_file))
    # Not uses set() here as we want to preserve the order!
    all_cdsl_sets = []
    instr_classes = {32}
    compressed = "c" in base_extensions
    instr_classes.add(16)
    for ext in base_extensions:
        if ext == "c":
            continue
        cdsl_sets = get_cdsl_sets(ext, xlen=xlen, compressed=compressed)
        for cdsl_set in cdsl_sets:
            if cdsl_set not in all_cdsl_sets:
                all_cdsl_sets.append(cdsl_set)
    if not ignore_etiss:
        all_cdsl_sets = apply_etiss_overrides(all_cdsl_sets, semihosting=semihosting)
    # print("all_cdsl_sets", all_cdsl_sets)
    core_includes = get_cdsl_includes(all_cdsl_sets, base_dir=base_dir, tum_dir=tum_dir)
    for set_name in set_names:
        set_out_cdsl_file = gen_dir / f"{set_name}.core_desc"
        core_includes.add(set_out_cdsl_file.resolve())
    # print("cdsl_includes", cdsl_includes)
    core_includes_code = get_includes_code(core_includes)
    constants = {}
    memories = {}
    constants["XLEN"] = m2isar.metamodel.arch.Constant("XLEN", value=xlen, attributes={}, size=None, signed=False)
    # if ignore_etiss:
    if True:
        main_reg = m2isar.metamodel.arch.Memory(
            "X",
            m2isar.metamodel.arch.RangeSpec(32),
            size=xlen,
            attributes={m2isar.metamodel.arch.MemoryAttribute.IS_MAIN_REG: []},
        )
        main_mem = m2isar.metamodel.arch.Memory(
            "MEM",
            m2isar.metamodel.arch.RangeSpec(1 << 32),
            size=8,
            attributes={m2isar.metamodel.arch.MemoryAttribute.IS_MAIN_MEM: []},
        )
        pc = m2isar.metamodel.arch.Memory(
            "PC",
            m2isar.metamodel.arch.RangeSpec(0),
            size=32,
            attributes={m2isar.metamodel.arch.MemoryAttribute.IS_PC: []},
        )
        memories["X"] = main_reg
        memories["MEM"] = main_mem
        memories["PC"] = pc
    functions = {}
    intrinsics = {}
    contributing_types = all_cdsl_sets
    if extra_includes is None:
        extra_includes = []
    # extra_includes = ["/work/git/students/cecil/etiss_arch_riscv/rv_base"]  # TODO: Do not hardcode
    # extra_includes = ["/mnt/wd8tb/Data/students_archive/cecil/etiss_arch_riscv/rv_base"]  # TODO: Do not hardcode
    # print("SSSS")
    # name_idx = 0
    errs_file = gen_dir / "errs.txt"
    errs = {}
    unencoded_instructions_per_set = defaultdict(dict)
    instructions_per_set = defaultdict(dict)
    with ProcessPoolExecutor(n_parallel) as pool:

        futures = []
        for set_name, set_name_, set_file, set_encoded_file in generated_sets:
            if auto_encoding:
                set_file_ = set_file
            else:
                assert set_encoded_file is not None, "cdsl_encoded artifact not found with auto_encoding disabled"
                set_file_ = set_encoded_file
            future = pool.submit(
                parse_generated_set,
                set_file_,
                set_name=set_name,
                skip_errors=skip_errors,
                extra_includes=extra_includes,
                add_mnemonic_prefix=add_mnemonic_prefix,
                auto_encoding=auto_encoding,
            )
            futures.append(future)
        for future in futures:
            # TODO: except failing?
            result = future.result()
            parsed_instructions, set_name, err_msg = result
            if parsed_instructions is None:
                assert err_msg is not None
                errs[(set_name, set_name_)] = err_msg
            # assert set_name not in unencoded_instructions_per_set, f"Duplicate set name: {set_name}"
            if auto_encoding:
                unencoded_instructions_per_set[set_name].update(parsed_instructions)
            else:
                instructions_per_set[set_name].update(parsed_instructions)

    if auto_encoding:
        for set_name, unencoded_instructions in unencoded_instructions_per_set.items():
            unencoded_instructions_ = list(unencoded_instructions.values())
            encoded_instructions_ = encode_instructions(unencoded_instructions_)
            # print("encoded_instructions_", encoded_instructions_, len(encoded_instructions_))
            encoded_instructions = {(instr_def.code, instr_def.mask): instr_def for instr_def in encoded_instructions_}
            # print("encoded_instructions", encoded_instructions, len(encoded_instructions))
            instructions_per_set[set_name] = encoded_instructions
    for set_name in instructions_per_set.keys():
        contributing_types.append(set_name)
    # instructions = {}
    # instructions.update(encoded_instructions)
    # input("1")
    generated_core = m2isar.metamodel.arch.CoreDef(
        name=core_name,
        contributing_types=contributing_types,
        template=None,  # type: ignore
        constants=constants,
        memories=memories,
        memory_aliases={},
        functions=functions,
        # instructions=instructions,
        instructions={},
        unencoded_instructions={},
        instr_classes=instr_classes,
        intrinsics=intrinsics,
    )
    # print("generated_core", generated_core)
    with open(core_out_model_file, "wb") as f:
        cores = {core_name: generated_core}
        model_obj = M2Model(M2_METAMODEL_VERSION, cores, {}, {})
        pickle.dump(model_obj, f)
    core_cdsl_code = gen_cdsl_code(model_obj, with_includes=False)
    with open(core_out_cdsl_file, "w") as f:
        f.write(core_includes_code)
        f.write("\n\n")
        f.write(core_cdsl_code)
    extension = [f"RV{xlen}I"]  # TODO: add include?
    set_includes = get_cdsl_includes(extension, base_dir=base_dir, tum_dir=tum_dir)
    set_includes_code = get_includes_code(set_includes)
    set_hls_includes = []  # TODO: fix duplicate includes
    set_hls_includes_code = get_includes_code(set_hls_includes)
    for set_name, instructions in instructions_per_set.items():
        set_out_model_file = gen_dir / f"{set_name}.m2isarmodel"
        set_splitted_out_model_file = gen_dir / f"{set_name}.splitted.m2isarmodel"
        set_out_cdsl_file = gen_dir / f"{set_name}.core_desc"
        set_hls_out_cdsl_file = gen_dir / f"{set_name}.hls.core_desc"
        set_splitted_out_cdsl_file = gen_dir / f"{set_name}.splitted.core_desc"
        generated_set = m2isar.metamodel.arch.InstructionSet(
            name=set_name,
            extension=extension,
            constants={},
            memories={},
            functions={},
            instructions=instructions_per_set[set_name],
            unencoded_instructions={},
        )
        # print("generated_set", generated_set)
        with open(set_out_model_file, "wb") as f:
            sets = {set_name: generated_set}
            model_obj = M2Model(M2_METAMODEL_VERSION, {}, sets, {})
            pickle.dump(model_obj, f)
        set_cdsl_code = gen_cdsl_code(model_obj, with_includes=False)
        with open(set_out_cdsl_file, "w") as f:
            f.write(set_includes_code)
            f.write("\n\n")
            f.write(set_cdsl_code)
        sets_splitted = {}
        for instr_def in generated_set.instructions.values():
            instr_name = instr_def.name
            temp_set_name = f"{set_name}{instr_name}single"
            temp_set = m2isar.metamodel.arch.InstructionSet(
                name=temp_set_name,
                extension=extension,
                constants={},
                memories={},
                functions={},
                instructions={(instr_def.mask, instr_def.code): instr_def},
                unencoded_instructions={},
            )
            sets_splitted[temp_set_name] = temp_set
            if auto_encoding:
                model_obj_single = M2Model(M2_METAMODEL_VERSION, {}, {temp_set_name: temp_set}, {})
                sub_name = set_instr2sub_name[(set_name, instr_name)]
                sub_files = sub2files[sub_name]
                base_cdsl_file = str(sub_files["cdsl"])
                base_cdsl_file_encoded = base_cdsl_file.replace(".core_desc", ".encoded.core_desc")
                set_single_out_model_file = base_cdsl_file_encoded.replace(".core_desc", ".m2isarmodel")
                set_single_out_cdsl_file = base_cdsl_file_encoded
                with open(set_single_out_model_file, "wb") as f:
                    pickle.dump(model_obj_single, f)
                extension2 = [f"RV{xlen}I"]  # TODO: add include?
                set_includes2 = get_cdsl_includes(extension2, base_dir=base_dir, tum_dir=tum_dir)
                set_includes_code2 = get_includes_code(set_includes2)
                set_single_cdsl_code = gen_cdsl_code(model_obj_single, with_includes=False)
                with open(set_single_out_cdsl_file, "w") as f:
                    f.write(set_includes_code2)
                    f.write("\n\n")
                    f.write(set_single_cdsl_code)
                sub2new_files[sub_name]["cdsl_encoded"] = base_cdsl_file_encoded
        group_set = m2isar.metamodel.arch.InstructionSet(
            name=set_name,
            extension=list(sets_splitted.keys()),
            constants={},
            memories={},
            functions={},
            instructions={},
            unencoded_instructions={},
        )
        sets_splitted[set_name] = group_set
        model_obj_splitted = M2Model(M2_METAMODEL_VERSION, {}, sets_splitted, {})
        with open(set_splitted_out_model_file, "wb") as f:
            pickle.dump(model_obj_splitted, f)
        set_splitted_cdsl_code = gen_cdsl_code(model_obj_splitted, with_includes=False)
        with open(set_splitted_out_cdsl_file, "w") as f:
            f.write(set_includes_code)
            f.write("\n\n")
            f.write(set_splitted_cdsl_code)
        model_obj.sets[set_name].memories["X"] = main_reg
        model_obj.sets[set_name].extension = []
        hls_unique_instr_names = len(set_names) > 1
        if hls_unique_instr_names:
            assert add_mnemonic_prefix
            for instr_def in model_obj.sets[set_name].instructions.values():
                instr_def.name = instr_def.mnemonic.upper().replace(".", "_")
        set_hls_cdsl_code = gen_cdsl_code(model_obj, with_includes=False, legacy=True)
        with open(set_hls_out_cdsl_file, "w") as f:
            f.write(set_hls_includes_code)
            f.write("\n\n")
            f.write(set_hls_cdsl_code)
        if len(errs) > 0:
            with open(errs_file, "w") as f:
                errs_text = "\n".join([f"{name}: {err}" for name, err in errs.items()])
                f.write(errs_text)
    if auto_encoding:
        assert len(index_files) == 1
        combined_index_file = index_files[0]
        inplace = True
        assert inplace
        new_index_data = index_data
        for i in range(len(index_data["candidates"])):
            sub_name = f"C{i}"
            new_artifacts = sub2new_files.get(sub_name)
            if new_artifacts:
                new_index_data["candidates"][i]["artifacts"].update(new_artifacts)
        with open(combined_index_file, "w") as f:
            yaml.dump(new_index_data, f)


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    generate_etiss_core(
        sess,
        force=args.force,
        workdir=args.workdir,
        gen_dir=args.gen_dir,
        index_files=args.index.split(";") if args.index is not None else None,
        core_name=args.core_name,
        set_names=args.set_name.split(";"),
        xlen=args.xlen,
        ignore_etiss=args.ignore_etiss,
        semihosting=args.semihosting,
        base_extensions=args.base_extensions.split(","),
        auto_encoding=args.auto_encoding,
        split=args.split,
        base_dir=args.base_dir,
        tum_dir=args.tum_dir,
        skip_errors=args.skip_errors,
        extra_includes=(args.extra_includes.split(";") if args.extra_includes is not None else None),
        add_mnemonic_prefix=args.add_mnemonic_prefix,
        n_parallel=args.parallel,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--gen-dir", type=str, default=None)
    parser.add_argument("--index", type=str, default=None)
    parser.add_argument("--core-name", type=str, default="IsaacCore")
    parser.add_argument("--set-name", type=str, default="XIsaac")
    parser.add_argument("--xlen", type=int, default=32)
    parser.add_argument("--ignore_etiss", action="store_true")
    parser.add_argument("--semihosting", action="store_true")
    parser.add_argument("--base-extensions", type=str, default="i,m,a,f,d,c,zifencei")
    parser.add_argument("--auto-encoding", action="store_true", default=True)
    parser.add_argument("--no-auto-encoding", dest="auto_encoding", action="store_false", default=True)
    parser.add_argument("--split", action="store_true")
    parser.add_argument("--base-dir", type=str, default="rv_base")
    parser.add_argument("--tum-dir", type=str, default=".")
    parser.add_argument("--parallel", type=int, default=multiprocessing.cpu_count())
    parser.add_argument("--extra-includes", type=str, default=None)  # semicolon separated
    parser.add_argument("--skip-errors", action="store_true")
    parser.add_argument("--add-mnemonic-prefix", action="store_true")

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse

import yaml


def annotate_score(
    index,
    output=None,
    inplace=False,
    filtered=False,
    filtered2=False,
    prelim=False,
    final=False,
    util_weight=1.0,
    enc_weight=1.0,
):

    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)
    candidates = combined_index_data["candidates"]
    # print("candidates", candidates)

    score_name = "score"

    if filtered:
        assert not (final or filtered2 or prelim)
        score_name = f"filtered_{score_name}"
    if filtered2:
        assert not (final or filtered or prelim)
        score_name = f"filtered2_{score_name}"
    if prelim:
        assert not (final or filtered2 or filtered)
        score_name = f"prelim_{score_name}"
    if final:
        assert not (prelim or filtered2 or filtered)
        score_name = f"final_{score_name}"

    for i, candidate in enumerate(candidates):
        # print("i", i)
        # name = candidate["properties"]["InstrName"]
        metrics = candidate.get("metrics", {})
        # print("metrics", metrics)
        # input("!")
        # runtime_reduction_rel = metrics["runtime_reduction_rel"]
        # code_size_reduction_rel = metrics["code_size_reduction_rel"]
        benefits = []
        costs = []
        util_score = metrics.get("util_score")
        if util_score is not None:
            assert util_score >= 0
            util_score = util_score * util_weight
            benefits.append(util_score)

        enc_weight = metrics.get("enc_weight")
        if enc_weight is not None:
            assert enc_weight >= 0
            enc_weight = enc_weight * enc_weight
            costs.append(enc_weight)

        assert len(benefits) > 0
        benefits_sum = sum(benefits)

        assert len(costs) > 0
        costs_sum = sum(costs)

        score = benefits_sum / costs_sum
        metrics[score_name] = score
        # print("metrics2", metrics)
        candidate["metrics"] = metrics

    if inplace:
        assert output is None
        out_file = index
    else:
        assert output is not None
        out_file = output

    combined_index_data["candidates"] = candidates

    with open(out_file, "w") as f:
        yaml.dump(combined_index_data, f)


def handle(args):
    annotate_score(
        args.index,
        output=args.output,
        inplace=args.inplace,
        filtered=args.filtered,
        filtered2=args.filtered2,
        prelim=args.prelim,
        final=args.final,
        util_weight=args.util_weight,
        enc_weight=args.enc_weight,
    )


def get_parser():
    parser = argparse.ArgumentParser(description="TODO")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument("--inplace", action="store_true", help="TODO")
    parser.add_argument("--filtered", action="store_true", help="TODO")
    parser.add_argument("--filtered2", action="store_true", help="TODO")
    parser.add_argument("--prelim", action="store_true", help="TODO")
    parser.add_argument("--final", action="store_true", help="TODO")
    # parser.add_argument("--runtime-weight", type=float, default=1.0, help="TODO")
    # parser.add_argument("--code-size-weight", type=float, default=1.0, help="TODO")
    parser.add_argument("--util-weight", type=float, default=1.0, help="TODO")
    parser.add_argument("--enc-weight", type=float, default=1.0, help="TODO")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import logging
import argparse
from pathlib import Path

import numpy as np
import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, TableArtifact, filter_artifacts


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


def collect_perf_metrics_seq(timing_df):
    stages = timing_df.columns
    temp_df = timing_df.copy()
    temp_df2 = pd.DataFrame(temp_df.values < temp_df[[stages[0]]].values, columns=stages)
    temp_df2.index = temp_df.index
    mask_df = temp_df2

    timing_df[mask_df] = np.nan
    avg_stages_per_instr = timing_df.notna().sum(axis=1).mean(axis=0)
    latencies_df = (timing_df[stages].max(axis=1) - timing_df[stages].min(axis=1)) + 1
    avg_latency = latencies_df.mean()
    median_latency = latencies_df.median()
    max_latency = latencies_df.max()
    num_stages = len(stages)
    total_instructions = len(timing_df)
    total_cycles = timing_df[stages].max().max() - timing_df[stages].min().min()
    stall_cycles = total_cycles - total_instructions
    stall_rate = stall_cycles / total_cycles
    cpi = total_cycles / total_instructions
    # TODO: minus latency of first instruction?
    metrics = {
        "num_stages": num_stages,
        "avg_stages_per_instr": avg_stages_per_instr,
        "total_cycles": total_cycles,
        "total_instructions": total_instructions,
        "stall_cycles": stall_cycles,
        "stall_rate": stall_rate,
        "cpi": cpi,
        "avg_latency": avg_latency,
        "median_latency": median_latency,
        "max_latency": max_latency,
    }

    if cpi > 10:
        print("metrics", metrics)
        print(timing_df)
    # print("metrics", metrics)
    return metrics


def collect_perf_metrics(
    timing_df,
):
    index_values = np.array(timing_df.index)
    index_values_ref = np.arange(index_values[0], index_values[0] + len(index_values))
    index_values_diff = index_values_ref - index_values
    counts = np.unique(index_values_diff, return_counts=True)
    all_metrics = []
    if len(counts[0]) > 1:
        seqs = []
        while len(index_values) > 0:
            print("remaining", len(index_values))
            print("index_values_diff", index_values_diff)
            until = np.argmax(index_values_diff < 0)
            print("until", until)
            if until == 0:
                print("until==0")
                index_values_this = index_values
                index_values = []
                seqs.append(index_values_this)
            else:
                index_values_this = index_values[:until]
                index_values = index_values[until:]
                index_values_diff = index_values_diff[until:]
                index_values_diff = index_values_diff - index_values_diff[0]
                seqs.append(index_values_this)
        for seq in seqs:
            seq_timing_df = timing_df.loc[seq]
            start_idx = seq_timing_df.index[0]
            end_idx = seq_timing_df.index[-1]
            metrics = collect_perf_metrics_seq(seq_timing_df)
            metrics["start_idx"] = start_idx
            metrics["end_idx"] = end_idx
            all_metrics.append(metrics)

    else:
        assert len(counts[0]) == 1
        assert counts[0][0] == 0
        metrics = collect_perf_metrics_seq(timing_df)
        all_metrics.append(metrics)
    metrics_df = pd.DataFrame(all_metrics)
    print(metrics_df)
    print(metrics_df[["total_instructions", "cpi"]])

    return metrics_df


def get_perf_metrics(
    sess: Session,
    force: bool = False,
):
    artifacts = sess.artifacts

    timing_trace_artifacts = filter_artifacts(
        artifacts, lambda x: x.flags & ArtifactFlag.TRACE and x.attrs.get("kind") == "timing_trace"
    )
    assert len(timing_trace_artifacts) == 1
    timing_trace_artifact = timing_trace_artifacts[0]
    assert timing_trace_artifact.attrs.get("simulator") in ["etiss_perf", "etiss"]

    metrics_df = collect_perf_metrics(
        timing_trace_artifact.df,
    )

    attrs = {
        "timing_trace": timing_trace_artifact.name,
        "kind": "metrics",
        "by": __name__,
    }

    perf_metrics_artifact = TableArtifact("perf_metrics", metrics_df, attrs=attrs)
    sess.add_artifact(perf_metrics_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    get_perf_metrics(sess, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--verbose", "-v", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from typing import Optional, Union
from pathlib import Path

import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import TableArtifact
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


COLS = ["Model", "Arch", "Run Instructions", "Run Instructions (rel.)"]
MEM_COLS = ["Model", "Arch", "Total ROM", "Total RAM", "ROM code", "ROM code (rel.)"]
COMMON_COLS = list(set(COLS) & set(MEM_COLS))


def compare_bench(
    sess: Session,
    report: Union[str, Path] = None,
    mem_report: Optional[Union[str, Path]] = None,
    force: bool = False,
):
    logger.info("Comparing bench reports...")
    COLS = ["Model", "Arch", "Run Instructions", "Run Instructions (rel.)"]
    MEM_COLS = [
        "Model",
        "Arch",
        "Total ROM",
        "Total RAM",
        "ROM code",
        "ROM code (rel.)",
    ]
    COMMON_COLS = list(set(COLS) & set(MEM_COLS))

    assert report is not None

    report_file = Path(report)
    assert report_file.is_file()
    report_df = pd.read_csv(report_file)[COLS]

    if mem_report:
        mem_report_file = Path(mem_report)
        assert mem_report_file.is_file()
        mem_report_df = pd.read_csv(mem_report_file)[MEM_COLS]
        report_df = report_df.merge(mem_report_df, on=COMMON_COLS)

    compare_df = report_df

    attrs = {}

    artifact = TableArtifact("compare_bench", compare_df, attrs=attrs)
    sess.add_artifact(artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    compare_bench(
        sess,
        report=args.report,
        mem_report=args.mem_report,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--report", required=True, help="Report CSV file")
    parser.add_argument("--mem-report", default=None, help="Memory report CSV file")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import pickle
import argparse
from pathlib import Path

import pandas as pd


# TODO: argparse (--mem, ...)


def print_memory_footprint(df):
    print("================")
    print("Memory Footprint")
    print("================")
    mem = pd.concat([df.dtypes, df.memory_usage(deep=True)], axis=1).sort_values(1, ascending=False)
    mem.rename(columns={0: "dtype", 1: "mem"}, inplace=True)
    total = mem["mem"].sum()
    print(mem)
    print("TOTAL:", total)


def handle(args):
    file = Path(args.file)
    try:
        if ".pkl" in file.name:
            data = pd.read_pickle(args.file, compression="infer")
        elif ".parquet" in file.name:
            data = pd.read_parquet(args.file, compression="infer")
    except Exception:
        # Falling back to non-pandas
        with open(args.file, "rb") as f:
            data = pickle.load(f)
    if not args.skip_print:
        print("Unpickled Data:")
        with pd.option_context(
            "display.min_rows",
            args.max_rows,
            "display.max_rows",
            args.max_rows,
            "display.min_rows",
            args.min_rows,
            "display.max_columns",
            args.max_columns,
            "display.width",
            None,
            "max_colwidth",
            150,
        ):
            print(
                data
                if (args.print_topk is None or not isinstance(data, (pd.DataFrame, pd.Series)))
                else data.iloc[: args.print_topk]
            )

    if isinstance(data, (pd.DataFrame, pd.Series)) or hasattr(data, "__len__"):
        print("Shape:", data.shape)
    else:
        print(f"len={len(data)}")

    if args.memory:
        assert isinstance(data, (pd.DataFrame, pd.Series)), "Memory footprint only available for pandas types"
        print_memory_footprint(data)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument("--skip-print", action="store_true")
    parser.add_argument("--memory", action="store_true")
    parser.add_argument("--print-topk", type=int, default=None)
    parser.add_argument("--max-rows", type=int, default=None)
    parser.add_argument("--min-rows", type=int, default=None)
    parser.add_argument("--max-columns", type=int, default=None)
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2024 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import sys
import itertools
import random
from math import ceil

import yaml
import argparse
from typing import Optional, Union, List
from pathlib import Path
from collections import defaultdict
import pandas as pd

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()

# SUPPORTED_STRATEGIES = ["best", "worst", "random"]
SUPPORTED_VARIANT_STRATEGIES = [
    "min_ii",
    "min_lat",
    "min_area",
    "balanced",
    "random",
    "all",
]
DEFAULT_STRATEGIES = ["all"]
# DEFAULT_STRATEGIES = ["min_ii(topk=1)", "min_lat(topk=1)", "min_area(topk=1)", "balanced", "random(n=10)", "all(limit=100)"]


def parse_strategy_string(strategy_str: str):
    """
    Example:
    min_ii(topk=2)
    balanced(alpha=1,beta=1,gamma=1)
    random(n=5)
    all(limit=100,shuffle=true)
    """
    pattern = r"(\w+)(?:\((.*)\))?"
    match = re.match(pattern, strategy_str.strip())
    if not match:
        raise ValueError(f"Invalid strategy format: {strategy_str}")

    name = match.group(1)
    args_str = match.group(2)

    kwargs = {}
    if args_str:
        for part in args_str.split(","):
            k, v = part.split("=")
            k = k.strip()
            v = v.strip()
            if v.lower() == "true":
                v = True
            elif v.lower() == "false":
                v = False
            else:
                try:
                    v = int(v)
                except ValueError:
                    try:
                        v = float(v)
                    except ValueError:
                        pass
            kwargs[k] = v

    return name, kwargs


def generate_variants(sg_schedules, strategy_strings):
    """
    Returns:
        List[Dict[sg_id -> sol_idx]]
    """

    parsed = [parse_strategy_string(s) for s in strategy_strings]

    # deterministic ordering
    sg_ids = sorted(sg_schedules.keys())

    all_variants = []
    seen = set()

    def variant_key(v):
        return tuple((sg, v[sg]) for sg in sg_ids)

    for strategy_name, kwargs in parsed:

        # ---------------------------------
        # ALL
        # ---------------------------------
        if strategy_name == "all":
            limit = kwargs.get("limit", None)
            shuffle = kwargs.get("shuffle", False)

            per_sg_indices = [list(range(len(sg_schedules[sg]))) for sg in sg_ids]

            product = list(itertools.product(*per_sg_indices))

            if shuffle:
                random.shuffle(product)

            if limit is not None:
                product = product[:limit]
            else:
                MAX_NUM_ALL = 200
                num = len(product)
                if num >= MAX_NUM_ALL:
                    raise RuntimeError(
                        f"Strategy all would generate {num} variants! Use all(limit={num}) to allow this explicitly."
                    )

            for local_idx, combo in enumerate(product):
                variant = {sg: sol for sg, sol in zip(sg_ids, combo)}
                key = variant_key(variant)
                if key not in seen:
                    seen.add(key)
                    all_variants.append((variant, f"all:{local_idx}"))

        # ---------------------------------
        # RANDOM
        # ---------------------------------
        elif strategy_name == "random":
            n = kwargs.get("n", 1)

            for i in range(n):
                variant = {}
                for sg in sg_ids:
                    sol = random.randrange(len(sg_schedules[sg]))
                    variant[sg] = sol

                key = variant_key(variant)
                if key not in seen:
                    seen.add(key)
                    all_variants.append((variant, f"random(n={n}):{i}"))

        # ---------------------------------
        # MIN-II / MIN-LAT / MIN-AREA
        # ---------------------------------
        elif strategy_name in ["min_ii", "min_lat", "min_area", "balanced"]:

            topk = kwargs.get("topk", 1)

            per_sg_topk = []

            for sg in sg_ids:
                scheds = sg_schedules[sg]

                if strategy_name == "min_ii":
                    ranked = sorted(enumerate(scheds), key=lambda x: x[1]["ii"])

                elif strategy_name == "min_lat":
                    ranked = sorted(enumerate(scheds), key=lambda x: sum(x[1]["lats"].values()))

                elif strategy_name == "min_area":
                    ranked = sorted(enumerate(scheds), key=lambda x: 123.0)  # TODO real area

                elif strategy_name == "balanced":
                    alpha = kwargs.get("alpha", 1.0)
                    beta = kwargs.get("beta", 1.0)
                    gamma = kwargs.get("gamma", 1.0)

                    def score(item):
                        _, sched = item
                        ii = sched["ii"]
                        lat = sum(sched["lats"].values())
                        area = 123.0  # TODO: replace
                        return alpha * ii + beta * lat + gamma * area

                    ranked = sorted(enumerate(scheds), key=score)

                per_sg_topk.append([idx for idx, _ in ranked[:topk]])

            product = list(itertools.product(*per_sg_topk))

            for local_idx, combo in enumerate(product):
                variant = {sg: sol for sg, sol in zip(sg_ids, combo)}
                key = variant_key(variant)
                if key not in seen:
                    seen.add(key)
                    all_variants.append((variant, f"{strategy_name}:{local_idx}"))

        else:
            raise NotImplementedError(f"Unknown strategy {strategy_name}")

    return all_variants


def run_fake_hls(
    sess: Session,
    set_name: str = "XIsaac",
    core: Optional[str] = None,
    strategies: Optional[List[str]] = None,
    index: Optional[Union[str, Path]] = None,
    workdir: Optional[Union[str, Path]] = None,
    label: Optional[str] = None,
    force: bool = False,
    verbose: bool = False,
    cleanup: bool = False,
):
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    # TODO: handle suffix
    out_dir = workdir / "local" / "fake_hls"
    out_dir.mkdir(exist_ok=True, parents=True)
    print("FAKE HLS")
    if strategies is None:
        # strategies = [DEFAULT_STRATEGY]
        strategies = DEFAULT_STRATEGIES
    # assert len(strategies) == 1
    strategy = strategies[0]
    assert core is not None
    assert set_name is not None
    assert index is not None
    index_file = Path(index)
    assert index_file.is_file()
    instr_schedules = defaultdict(list)
    sg_schedules = defaultdict(list)
    with open(index_file, "r") as f:
        index_data = yaml.safe_load(f)
    candidates = index_data["candidates"]
    num_candidates = len(candidates)
    print("candidates", candidates, len(candidates))
    sharing_groups = defaultdict(list)
    first_stage = 2
    for i, candidate in enumerate(candidates):
        print("candidate", candidate)
        properties = candidate["properties"]
        metrics = candidate["metrics"]
        name = properties["InstrName"]
        # TODO: handle actual sharing
        sg = i + 1
        sg_ = [name]
        sharing_groups[sg] = sg_

        num_instrs = int(properties["#Instrs"])
        max_path_len = int(properties["ScheduleLength"])

        assert num_instrs >= 2
        min_lat = 1
        max_lat = num_instrs
        lats = list(range(min_lat, max_lat + 1))
        print("name", name)
        print("num_instrs", num_instrs)
        print("lats", lats)
        for lat in lats:
            print("lat", lat)
            min_ii = 1
            # max_ii = 1
            max_ii = lat
            iis = list(range(min_ii, max_ii + 1))
            print("iis", iis)
            for ii in iis:
                # stages should have the same latency? (can be maybe dropped)
                legal = (lat % ii) == 0
                if not legal:
                    continue
                # Filter candidates with unsupported timing model
                max_stages = 3
                required_stages = ceil(lat / ii)
                legal2 = required_stages <= max_stages
                if not legal2:
                    continue
                full_lat = lat + first_stage
                sched = {"lat": lat, "full_lat": full_lat, "ii": ii}
                instr_schedules[name].append(sched)
                sg_sched = {"lats": {name: lat}, "full_lats": {name: full_lat}, "ii": ii}
                sg_schedules[sg].append(sg_sched)

    print("instr_schedules", instr_schedules)
    print("sg_schedules", sg_schedules)
    # strategies = ["fast", "slow"]
    # strategy = "worst"  # TODO: implement others
    # selected_schedules = defaultdict(dict)
    # selected_schedules_idx = defaultdict(int)
    print("sharing_groups", sharing_groups)

    variants = generate_variants(sg_schedules, strategies)
    # print("variants", variants, len(variants))
    # input("!")
    # print("selected_schedules", selected_schedules)
    # print("selected_schedules_idx", selected_schedules_idx)
    # print("selected_sg_schedules", selected_sg_schedules)
    # print("selected_sg_schedules_idx", selected_sg_schedules_idx)
    # TODO: allow sharing
    # TODO: analyze nodes (find longes paths,...)
    # TODO: use external cost/latency model
    # TODO: use constraints
    # TODO: sample schedule-combinations
    # Output format
    hls_schedules_csv_data = []
    all_hls_schedules_csv_data = []
    max_stage = first_stage
    # for instr_name, scheds in hls_schedules.items():
    #     for sol_idx, sched in enumerate(scheds):
    #         config = f"SG_{sg}_SOL_IDX_{sol_idx}"
    #         ii = sched["ii"]
    #         lat = sched["lat"]
    #         full_lat = lat + first_stage
    #         lats = {instr_name: full_lat}
    #         allocs = {}
    #         area_est = 123.0  # DUMMY
    #         area_est2 = area_est
    #         new2 = {
    #             "config": config,
    #             "idx": sol_idx,
    #             "II": ii,
    #             "Fallback": False,
    #             "Instruction latencies": lats,
    #             "Allocation": allocs,
    #             "Overall latency": max_lat,
    #             "Area estimate w/o lifetimes": area_est,
    #             "Area estimate w/ lifetimes": area_est2,
    #             "Total lifetime": 0.0,
    #             "Total decoupled ops": 0,
    #         }
    #         hls_schedules_csv_data.append(new2)
    #         sg += 1
    # TODO: fill
    default_area = 1.0
    default_delay = 1.0  # TODO
    default_sharing_weight = 0.2  # TODO
    instr_cost_dict = {
        "ADD": {
            "*": (default_area, default_delay, default_sharing_weight),
            "i32": (default_area, default_delay, default_sharing_weight),
            "i64": (default_area, default_delay, default_sharing_weight),
        },
        "ADDI": {
            "*": (default_area, default_delay, default_sharing_weight),
            "i32": (default_area, default_delay, default_sharing_weight),
            "i64": (default_area, default_delay, default_sharing_weight),
        },
        "SLL": {
            "*": (default_area, default_delay, default_sharing_weight),
            "i32": (default_area, default_delay, default_sharing_weight),
            "i64": (default_area, default_delay, default_sharing_weight),
        },
        "SLLI": {
            "*": (default_area, default_delay, default_sharing_weight),
            "i32": (default_area, default_delay, default_sharing_weight),
            "i64": (default_area, default_delay, default_sharing_weight),
        },
        "*": {
            "*": (default_area, default_delay, default_sharing_weight),
            "i32": (default_area, default_delay, default_sharing_weight),
            "i64": (default_area, default_delay, default_sharing_weight),
        },
    }

    def get_estimated_area(sg_sched, instr_cost_dict):
        instrs = list(sg_sched["lats"].keys())
        total_area = 0
        for instr in instrs:
            dtype = "unknown"
            lookup_key = instr if instr in instr_cost_dict else "*"
            assert lookup_key in instr_cost_dict
            dtype_cost_dict = instr_cost_dict[lookup_key]
            lookup_key2 = dtype if dtype in dtype_cost_dict else "*"
            assert lookup_key2 in dtype_cost_dict
            cost = dtype_cost_dict[lookup_key2]
            area, _, _ = cost
            total_area += area
        return total_area

    for sg, sg_scheds in sg_schedules.items():
        for sol_idx, sg_sched in enumerate(sg_scheds):
            config = f"SG_{sg}_SOL_IDX_{sol_idx}"
            ii = sg_sched["ii"]
            full_lats = sg_sched["full_lats"]
            allocs = {}
            area_est = get_estimated_area(sg_sched, instr_cost_dict)  # WIP
            area_est2 = area_est
            new2 = {
                "config": config,
                "idx": sol_idx,
                "II": ii,
                "Fallback": False,
                "Instruction latencies": full_lats,
                "Allocation": allocs,
                "Overall latency": max_lat,
                "Area estimate w/o lifetimes": area_est,
                "Area estimate w/ lifetimes": area_est2,
                "Total lifetime": 0.0,
                "Total decoupled ops": 0,
            }
            hls_schedules_csv_data.append(new2)
    # for instr_name, sched in selected_schedules.items():
    #     sol_idx = selected_schedules_idx[instr_name]
    #     new = {"sharing_group": sg, "solution_idx": sol_idx}
    #     selected_solutions_yaml_data.append(new)
    #     config = f"SG_{sg}_SOL_IDX_{sol_idx}"
    #     ii = sched["ii"]
    #     lat = sched["lat"]
    #     full_lat = lat + first_stage
    #     lats = {instr_name: full_lat}
    #     allocs = {}
    #     area_est = 123.0  # DUMMY
    #     area_est2 = area_est
    #     new2 = {
    #         "config": config,
    #         "idx": sol_idx,
    #         "II": ii,
    #         "Fallback": False,
    #         "Instruction latencies": lats,
    #         "Allocation": allocs,
    #         "Overall latency": max_lat,
    #         "Area estimate w/o lifetimes": area_est,
    #         "Area estimate w/ lifetimes": area_est2,
    #         "Total lifetime": 0.0,
    #         "Total decoupled ops": 0,
    #     }
    #     hls_schedules_csv_data.append(new2)
    #     stage = first_stage + lat - 1
    #     max_stage = max(max_stage, stage)
    #     dummy_sched = [{"interface": "foo", "stage": first_stage}, {"interface": "bar", "stage": stage}]
    #     new3 = {"instruction": instr_name, "schedule": dummy_sched}
    #     isax_xisaac_yaml_data.append(new3)
    #     sg += 1
    # print("hls_schedules_csv_data", hls_schedules_csv_data)
    # print("isax_xisaac_yaml_data", isax_xisaac_yaml_data)
    hls_schedules_csv_path = out_dir / "hls_schedules.csv"
    hls_outputs_path = out_dir / "output"
    hls_outputs_path.mkdir(exist_ok=True)
    hls_schedules_df = pd.DataFrame(hls_schedules_csv_data)
    hls_schedules_df.to_csv(hls_schedules_csv_path)

    # total_area_estimate = 0
    # total_area_estimate_with_lifetimes = 0
    # iis = []
    # all_lats = []
    # num_groups = 0
    # num_instrs = 0
    # group2instrs = {}
    # for row in selected_solutions_yaml_data:
    #     num_groups += 1
    #     sharing_group = row["sharing_group"]
    #     idx = row["solution_idx"]
    #     name = f"SG_{sharing_group}_SOL_IDX_{idx}"
    #     schedules = hls_schedules_df[hls_schedules_df["config"] == name]
    #     assert len(schedules) == 1
    #     # print("schedules", schedules)
    #     schedule = schedules.iloc[0]
    #     ii = schedule["II"]
    #     iis.append(ii)
    #     lats = schedule["Instruction latencies"]
    #     if isinstance(lats, str):
    #         lats = ast.literal_eval(lats)
    #     assert isinstance(lats, dict)
    #     group2instrs[idx] = list(lats.keys())
    #     num_instrs += len(lats)
    #     all_lats += list(lats.values())
    #     area_estimate = schedule["Area estimate w/o lifetimes"]
    #     total_area_estimate += area_estimate
    #     area_estimate_with_lifetimes = schedule["Area estimate w/ lifetimes"]
    #     total_area_estimate_with_lifetimes += area_estimate_with_lifetimes
    # max_instrs = max(map(len, group2instrs.values()))
    # min_instrs = min(map(len, group2instrs.values()))
    # avg_instrs = num_instrs / num_groups
    # min_ii = min(iis)
    # max_ii = max(iis)
    # avg_ii = sum(iis) / len(iis)
    # min_lat = min(all_lats)
    # max_lat = max(all_lats)
    # avg_lat = sum(all_lats) / len(all_lats)
    # hls_selected_schedule_metrics_data = [
    #     {
    #         "num_groups": num_groups,
    #         "num_instrs": num_instrs,
    #         "max_instrs": max_instrs,
    #         "min_instrs": min_instrs,
    #         "avg_instrs": avg_instrs,
    #         "min_ii": min_ii,
    #         "max_ii": max_ii,
    #         "avg_ii": avg_ii,
    #         "min_lat": min_lat,
    #         "max_lat": max_lat,
    #         "avg_lat": avg_lat,
    #         "total_area_estimate": total_area_estimate,
    #         "total_area_estimate_with_lifetimes": total_area_estimate_with_lifetimes,
    #     }
    # ]
    # hls_selected_schedule_metrics_df = pd.DataFrame([hls_selected_schedule_metrics_data])
    # print("hls_selected_schedule_metrics_df", hls_selected_schedule_metrics_df)
    hls_selected_schedule_metrics_csv_path = out_dir / "hls_selected_schedule_metrics.csv"
    # hls_selected_schedule_metrics_df.to_csv(hls_selected_schedule_metrics_csv_path)

    def process_variant(variant_ifx, variant_name, variant, description, variant_dir):
        variant_dir.mkdir(exist_ok=True)

        variant_selected_solutions_yaml_data = []
        variant_isax_xisaac_yaml_data = []

        total_area_estimate = 0
        total_area_estimate_with_lifetimes = 0
        iis = []
        all_lats = []
        num_groups = 0
        num_instrs = 0
        group_instr_counts = []
        details = []
        max_stage = 0

        for sg, sol_idx in variant.items():
            num_groups += 1

            sg_sched = sg_schedules[sg][sol_idx]

            ii = sg_sched["ii"]
            iis.append(ii)

            lats = sg_sched["lats"]
            full_lats = sg_sched["full_lats"]
            instr_count = len(full_lats)
            group_instr_counts.append(instr_count)
            num_instrs += instr_count
            all_lats += list(full_lats.values())

            area_estimate = 123.0
            total_area_estimate += area_estimate
            total_area_estimate_with_lifetimes += area_estimate
            new = {"sharing_group": sg, "solution_idx": sol_idx}
            variant_selected_solutions_yaml_data.append(new)
            for instr_name, lat in lats.items():
                stage = first_stage + lat - 1  # TODO!
                max_stage = max(max_stage, stage)
                dummy_sched = [{"interface": "foo", "stage": first_stage}, {"interface": "bar", "stage": stage}]
                new3 = {"instruction": instr_name, "schedule": dummy_sched}
                variant_isax_xisaac_yaml_data.append(new3)
            lats_str = ", ".join(f"{instr}: {lat}" for instr, lat in lats.items())
            full_lats_str = ", ".join(f"{instr}: {lat}" for instr, lat in full_lats.items())
            detail = f"SG{sg}(II={ii}, lats={{{lats_str}}}, full_lats={{{full_lats_str}}})"
            details.append(detail)
        details_str = ", ".join(details)
        # print("details_str", details_str)
        # input("!")
        # variant_isax_xisaac_yaml_data.append({"last_stage": max_stage + 1})
        variant_isax_xisaac_yaml_data.append({"last stage": max_stage + 1})
        variant_selected_solutions_yaml_path = variant_dir / "selected_solutions.yaml"
        # print("variant_selected_solutions_yaml_data", variant_selected_solutions_yaml_data)
        with open(variant_selected_solutions_yaml_path, "w") as f:
            yaml.dump(variant_selected_solutions_yaml_data, f)
        variant_isax_xisaac_yaml_path = variant_dir / "ISAX_XIsaac.yaml"
        with open(variant_isax_xisaac_yaml_path, "w") as f:
            yaml.dump(variant_isax_xisaac_yaml_data, f)

        max_instrs = max(group_instr_counts)
        min_instrs = min(group_instr_counts)
        avg_instrs = num_instrs / num_groups

        min_ii = min(iis)
        max_ii = max(iis)
        avg_ii = sum(iis) / len(iis)

        min_lat = min(all_lats)
        max_lat = max(all_lats)
        avg_lat = sum(all_lats) / len(all_lats)

        variant_metrics_row = {
            "num_groups": num_groups,
            "num_instrs": num_instrs,
            "max_instrs": max_instrs,
            "min_instrs": min_instrs,
            "avg_instrs": avg_instrs,
            "min_ii": min_ii,
            "max_ii": max_ii,
            "avg_ii": avg_ii,
            "min_lat": min_lat,
            "max_lat": max_lat,
            "avg_lat": avg_lat,
            "total_area_estimate": total_area_estimate,
            "total_area_estimate_with_lifetimes": total_area_estimate_with_lifetimes,
            "Variant idx": variant_idx,
            "Variant name": variant_name,
            "Variant description": description,
            "Variant details": details_str,
        }
        return variant_metrics_row

    variant_metrics_rows = []

    variants_dir = hls_outputs_path / "variants"
    for variant_idx, (variant, description) in enumerate(variants):
        variant_name = f"V{variant_idx}"
        variant_dir = hls_outputs_path / variant_name
        variant_metrics_row = process_variant(variant_idx, variant_name, variant, description, variant_dir)
        variant_metrics_rows.append(variant_metrics_row)
        if variant_idx == 0:
            # export first variant to base out dir for annotation (assign_hls)
            _ = process_variant(variant_idx, variant_name, variant, description, hls_outputs_path)
    hls_selected_schedule_metrics_df = pd.DataFrame(variant_metrics_rows)
    # print("hls_selected_schedule_metrics_df")
    # print(hls_selected_schedule_metrics_df)
    hls_selected_schedule_metrics_df.to_csv(hls_selected_schedule_metrics_csv_path)


def handle(args):
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    # strategy = args.strategy
    strategies = args.strategies
    # assert strategy is not None or strategies is not None
    # assert strategy is not None or strategies is not None
    # if strategy is not None:
    #     assert strategies is None
    #     strategies = [strategy]
    if strategies is not None:
        # assert strategy is None
        assert isinstance(strategies, str)
        strategies = strategies.split(",")
        # assert all(x in SUPPORTED_STRATEGIES for x in strategies)
    run_fake_hls(
        sess,
        set_name=args.set_name,
        core=args.core,
        strategies=strategies,
        index=args.index,
        force=args.force,
        workdir=args.workdir,
        verbose=args.verbose,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--set-name", type=str, default=None)
    parser.add_argument("--index", type=str, default=None)
    parser.add_argument("--core", type=str, choices=["cv32e40p"], default=None)
    # parser.add_argument("--strategy", type=str, choices=SUPPORTED_STRATEGIES, default=None)  # TODO: drop
    # parser.add_argument("--strategy", type=str, default=None)  # TODO: drop
    parser.add_argument("--strategies", type=str, default=None)
    parser.add_argument("--verbose", action="store_true")

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
import subprocess
import multiprocessing
from typing import Optional, Union
from pathlib import Path


from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.utils.annotate_enc_score import annotate_enc_score
from isaac_toolkit.utils.analyze_encoding import analyze_encoding

logger = get_logger()


def generate_cdsl(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    index_file: Optional[Union[str, Path]] = None,
    gen_dir: Optional[Union[str, Path]] = None,
    n_parallel: Optional[int] = None,
    suffix: str = "",
    force: bool = False,
):
    del sess, force
    logger.info("Generating CDSL...")
    if workdir is not None:
        workdir = Path(workdir)
    combined_index_file = workdir / "combined_index.yml" if index_file is None else Path(index_file)
    assert combined_index_file.is_file()
    # with open(combined_index_file, "r") as f:
    #     index_data = yaml.safe_load(f)

    gen_dir = Path(gen_dir) if gen_dir is not None else workdir / "gen"
    gen_dir.mkdir(exist_ok=True)
    generate_args = [
        combined_index_file,
        "--output",
        gen_dir,
        "--split",
        "--split-files",
        "--progress",
        "--inplace",  # TODO use gen/index.yml instead!
    ]
    if n_parallel is not None:
        generate_args += ["--parallel", str(n_parallel)]
    generate_cdsl_args = [
        "python3",
        "-m",
        "tool.gen.cdsl",
        *generate_args,
    ]
    generate_flat_args = [
        "python3",
        "-m",
        "tool.gen.flat",
        *generate_args,
    ]
    generate_fuse_cdsl_args = [
        "python3",
        "-m",
        "tool.gen.fuse_cdsl",
        *generate_args,
    ]
    # print("generate_cdsl_args", generate_cdsl_args)
    # print("generate_flat_args", generate_flat_args)
    # print("generate_fuse_cdsl_args", generate_fuse_cdsl_args)
    subprocess.run(generate_cdsl_args, check=True)
    subprocess.run(generate_flat_args, check=True)
    subprocess.run(generate_fuse_cdsl_args, check=True)
    enc_score_csv = workdir / f"encoding_score{suffix}.csv"
    total_enc_metrics_csv = workdir / f"total_encoding_metrics{suffix}.csv"
    analyze_encoding(combined_index_file, score=enc_score_csv, output=total_enc_metrics_csv)
    annotate_enc_score(combined_index_file, inplace=True, enc_score_csv=enc_score_csv)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    generate_cdsl(
        sess,
        workdir=args.workdir,
        gen_dir=args.gen_dir,
        index_file=args.index,
        force=args.force,
        n_parallel=args.parallel,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--gen-dir", type=str, default=None)
    parser.add_argument("--index", type=str, default=None)
    parser.add_argument("--parallel", type=int, default=multiprocessing.cpu_count())
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# import time
import re
import sys
import pandas as pd
import argparse
import logging
from pathlib import Path

from tqdm import tqdm

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import InstrTraceArtifact, TableArtifact


# TODO: logger
logger = logging.getLogger(__name__)


def load_riscvovpsimplus_trace(
    sess,
    input_file: Path,
    force: bool = False,
    progress: bool = False,
    chunksize: int = 2**20,
):
    logger.info("Loading riscvOVPSimPlus instruction trace...")
    assert input_file.is_file()
    name = input_file.name

    instr_chunks = []
    reg_chunks = []
    mem_chunks = []

    instr_idx = 0  # running instruction index

    def flush_chunks():
        """Convert buffers into DataFrames for concatenation."""
        df_instr = pd.DataFrame(instr_chunks)
        df_reg = pd.DataFrame(reg_chunks)
        df_mem = pd.DataFrame(mem_chunks)

        return df_instr, df_reg, df_mem

    with open(input_file, "r") as f:
        buffer_instr = []
        buffer_reg = []
        buffer_mem = []

        for line in tqdm(f, disable=not progress):
            line = line.strip()
            if not line.startswith("Info"):
                continue
            # print("line0", line)
            # Instruction line
            if re.match(r"Info '.*cpu.*',", line):
                m = re.match(
                    r"Info '.*',\s*(0x[0-9a-fA-F]+)\((.*?)\):\s+([0-9a-fA-F]+)\s+(\S+)(?:\s+(.*))?",
                    line,
                )
                if m:
                    pc, symbol, bytecode, instr, args = m.groups()
                    buffer_instr.append(
                        {
                            # "idx": instr_idx,
                            "pc": int(pc, 16),
                            "symbol": symbol,
                            "bytecode": int(bytecode, 16),
                            "instr": instr,
                            "args": args if args else "",
                            "size": 2 if instr.startswith("c.") or len(bytecode) <= 4 else 4,
                        }
                    )
                    instr_idx += 1

            # Register update line
            elif re.match(r"Info\s+[a-zA-Z0-9]+ ", line):
                print("line1", line)
                m = re.match(r"Info\s+(\w+)\s+([0-9a-fA-F]+)\s*->\s*([0-9a-fA-F]+)", line)
                if m:
                    reg, old_val, new_val = m.groups()
                    buffer_reg.append(
                        {
                            "idx": instr_idx - 1,  # belongs to last instr
                            "reg": reg,
                            "old_val": int(old_val, 16),
                            "new_val": int(new_val, 16),
                        }
                    )

                m = re.match(
                    r"Info\s+(MEM\w)\s+(0x[0-9a-fA-F]+)\s+(0x[0-9a-fA-F]+)\s+(\d+)\s+([0-9a-fA-F]+)",
                    line,
                )

                # if "MEM" in line:
                #    # print("line", line)
                #    print("m", m)
                if m:
                    acc_type, addr_lo, addr_hi, size, data = m.groups()
                    buffer_mem.append(
                        {
                            "idx": instr_idx - 1,  # belongs to last instr
                            "type": acc_type,
                            "addr_lo": int(addr_lo, 16),
                            "addr_hi": int(addr_hi, 16),
                            "size": int(size),
                            "data": int(data, 16),
                        }
                    )

            # Flush to disk in batches
            if len(buffer_instr) >= chunksize:
                instr_chunks.extend(buffer_instr)
                reg_chunks.extend(buffer_reg)
                mem_chunks.extend(buffer_mem)
                buffer_instr, buffer_reg, buffer_mem = [], [], []

        # final flush
        instr_chunks.extend(buffer_instr)
        reg_chunks.extend(buffer_reg)
        mem_chunks.extend(buffer_mem)

    # Build DataFrames
    df_instr = pd.DataFrame(instr_chunks)
    if not df_instr.empty:
        df_instr["instr"] = df_instr["instr"].astype("category")
        df_instr["size"] = df_instr["size"].astype("category")

    df_reg = pd.DataFrame(reg_chunks)
    df_mem = pd.DataFrame(mem_chunks)

    # Wrap artifacts
    attrs = {
        "simulator": "riscvOVPSimPlus",
        "cpu_arch": "unknown",
        "by": "isaac_toolkit.frontend.instr_trace.riscvovpsimplus",
    }
    instr_artifact = InstrTraceArtifact(name, df_instr, attrs=attrs)
    sess.add_artifact(instr_artifact, override=force)

    if not df_reg.empty:
        sess.add_artifact(TableArtifact(name + "_reg_updates", df_reg), override=force)
    if not df_mem.empty:
        sess.add_artifact(TableArtifact(name + "_mem_updates", df_mem), override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    input_file = Path(args.file)
    load_riscvovpsimplus_trace(sess, input_file, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--operands", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2024 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sys
import shutil
import subprocess

# import yaml
import argparse
from typing import Optional, Union, List
from pathlib import Path

from isaac_toolkit.session import Session
from isaac_toolkit.logging import get_logger, set_log_level
from isaac_toolkit.utils.seal5_score import calc_seal5_score

logger = get_logger()

DEFAULT_DOCKER_IMAGE = "isaac-quickstart-seal5:latest"


def retarget_seal5_llvm(
    sess: Session,
    workdir: Optional[Union[str, Path]] = None,
    mount_dir: Optional[Union[str, Path]] = None,
    docker_image: Optional[str] = None,
    seal5_sets: Optional[List[str]] = None,
    name: Optional[str] = None,
    label: Optional[str] = None,
    cfg_files: Optional[List[Union[str, Path]]] = None,
    splitted: bool = False,
    xlen: Optional[int] = 32,
    force: bool = False,
    verbose: bool = False,
    cleanup: bool = False,
    progress: bool = False,
):
    logger.info("Retargeting Seal5 LLVM...")
    assert xlen == 32
    assert workdir is not None
    if not isinstance(workdir, Path):
        workdir = Path(workdir)
    assert workdir.is_dir()
    if seal5_sets is None:
        seal5_sets = ["XIsaac"]
    print(f"label '{label}'")
    if label is None:
        # label = "splitted" if splitted else ""
        label = ""
    if name is None:
        name = "seal5_splitted" if splitted else "seal5"
    assert cfg_files is not None
    assert len(cfg_files) > 0
    use_docker = docker_image is not None
    subdir = "docker" if use_docker else "local"
    base_dir = workdir / subdir
    # seal5_dir = base_dir / "seal5"
    # output_dir = seal5_dir / label
    output_dir = (base_dir / name) if label == "" else (base_dir / f"{name}_{label}")
    if output_dir.is_dir():
        assert force, f"Directory already exists: {output_dir}. Use --force or different --label."
        logger.info("Cleaning up old output dir: %s (--force)", output_dir)
        shutil.rmtree(output_dir)
    output_dir.mkdir(parents=True)
    gen_dir = workdir / "gen"
    # gen_dir = workdir / "gen" / label
    gen_dir = (workdir / "gen") if label == "" else (workdir / f"gen_{label}")
    cdsl_files = [
        (gen_dir / f"{set_name}.splitted.core_desc" if splitted else gen_dir / f"{set_name}.core_desc")
        for set_name in seal5_sets
    ]
    print("cdsl_files", cdsl_files)
    print("cfg_files", cfg_files)
    # input("?")
    # TODO: PROGRESS
    if use_docker:
        command = "docker run -i --rm"
        if mount_dir is not None:
            command += f" -v {mount_dir}:{mount_dir}"
        command += f" -e CLEANUP={int(cleanup)}"
        command += f" {docker_image}"
        command += f" {output_dir}"
        command += " "
        command += " ".join(map(lambda x: str(Path(x).resolve()), cdsl_files))
        command += " "
        command += " ".join(map(lambda x: str(Path(x).resolve()), cfg_files))

        # print("$$$", command)
        kwargs = {}
        # print("verbose", verbose)
        if not verbose:
            kwargs.setdefault("stdout", subprocess.PIPE)
            kwargs.setdefault("stderr", subprocess.PIPE)
            kwargs.setdefault("text", True)
        try:
            subprocess.run(command, check=True, shell=True, **kwargs)
        except subprocess.CalledProcessError as e:
            print(f"[ERROR] Command failed with return code {e.returncode}")
            if e.stdout:
                print("--- STDOUT ---")
                print(e.stdout)
            if e.stderr:
                print("--- STDERR ---")
                print(e.stderr)
            raise  # Re-raise if you want the caller to handle it too
    else:
        seal5_script_local = os.environ.get("SEAL5_SCRIPT_LOCAL", None)
        assert seal5_script_local is not None, "Undefined: SEAL5_SCRIPT_LOCAL"
        args = [seal5_script_local]
        env = os.environ.copy()
        env["CLEANUP"] = str(int(cleanup))
        env["MGCLIENT_ROOT"] = env["MGCLIENT_INSTALL_DIR"]
        enable_cdfg_pass = True
        env["ENABLE_CDFG_PASS"] = str(int(enable_cdfg_pass))
        temp_dir = base_dir / "temp"
        temp_seal5_home = temp_dir / "seal5_llvm"
        env["SEAL5_HOME"] = temp_seal5_home
        ccache = True  # TODO: expose
        env["CCACHE"] = str(int(ccache))
        if ccache:
            ccache_dir = env.get("CCACHE_DIR", None)
            assert ccache_dir is not None, "Undefined: CCACHE_DIR"
            env["CCACHE_DIR"] = ccache_dir
        seal5_cfg_dir = env.get("SEAL5_CFG_DIR", None)
        assert seal5_cfg_dir is not None, "Undefined: SEAL5_CFG_DIR"
        env["SEAL5_CFG_DIR"] = seal5_cfg_dir
        seal5_dir = env.get("SEAL5_DIR", None)
        assert seal5_dir is not None, "Undefined: SEAL5_DIR"
        env["SEAL5_DIR"] = seal5_dir
        llvm_dir = env.get("LLVM_DIR", None)
        assert llvm_dir is not None, "Undefined: LLVM_DIR"
        print("llvm_dir", llvm_dir)
        env["LLVM_REPO"] = llvm_dir
        llvm_ref = "isaacnew-base-3"  # TODO: expose
        env["LLVM_REF"] = llvm_ref
        clone_depth = 2  # TODO: expose?
        env["CLONE_DEPTH"] = str(clone_depth)
        args += [output_dir]
        args += list(map(lambda x: str(Path(x).resolve()), cdsl_files))
        args += list(map(lambda x: str(Path(x).resolve()), cfg_files))
        # print("env", env)
        # print("$$$", args)
        kwargs = {}
        # print("verbose", verbose)
        if not verbose:
            kwargs.setdefault("stdout", subprocess.PIPE)
            kwargs.setdefault("stderr", subprocess.PIPE)
            kwargs.setdefault("text", True)
        try:
            subprocess.run(args, check=True, env=env, **kwargs)
        except subprocess.CalledProcessError as e:
            print(f"[ERROR] Command failed with return code {e.returncode}")
            if e.stdout:
                print("--- STDOUT ---")
                print(e.stdout)
            if e.stderr:
                print("--- STDERR ---")
                print(e.stderr)
            raise  # Re-raise if you want the caller to handle it too
    seal5_reports_dir = output_dir / "seal5_reports"
    seal5_status_csv = seal5_reports_dir / "status.csv"
    seal5_status_compact_csv = seal5_reports_dir / "status_compact.csv"
    seal5_score_csv = output_dir / "seal5_score.csv"
    calc_seal5_score(seal5_status_csv, seal5_status_compact_csv, output=seal5_score_csv)


def handle(args):
    # assert args.session is not None
    sess = None
    if args.session is not None:
        session_dir = Path(args.session)
        assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
        sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    retarget_seal5_llvm(
        sess,
        force=args.force,
        workdir=args.workdir,
        docker_image=args.docker,
        verbose=args.verbose,
    )
    if sess is not None:
        sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    # parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--session", "--sess", "-s", type=str, required=False)
    parser.add_argument("--force", "-f", action="store_true")
    parser.add_argument("--docker", type=str, default=None, const=DEFAULT_DOCKER_IMAGE, nargs="?")
    parser.add_argument("--workdir", type=str, default=None)
    parser.add_argument("--verbose", action="store_true")
    # label: Optional[str] = None,
    # etiss_core: Optional[str] = None,

    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import argparse
from pathlib import Path

from neo4j import GraphDatabase, Query
import networkx as nx

# import matplotlib.pyplot as plt

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import GraphArtifact
from isaac_toolkit.logging import get_logger, set_log_level

logger = get_logger()


def legalize_str(x):
    # logger.debug("legalize_str", x, type(x))
    legalized = x.replace("/", "_")
    # logger.debug("legalized", legalized)
    return legalized


def get_cfg_artifacts(driver, label: str = "default"):
    query = Query(
        f"""
    MATCH (n)-[r:CFG]->(c)
    WHERE n.session = "{label}"
    RETURN *
    """
    )

    session = driver.session()
    try:
        results = session.run(query)
        # logger.debug("results", results)

        G = nx.MultiDiGraph()

        nodes = list(results.graph()._nodes.values())
        # logger.debug("nodes", nodes, len(nodes))
        module_func_nodes = {}
        for node in nodes:
            props = node._properties
            module_name = props["module_name"]
            func_name = props["func_name"]
            if module_name not in module_func_nodes:
                module_func_nodes[module_name] = {}
            if func_name not in module_func_nodes[module_name]:
                module_func_nodes[module_name][func_name] = set()
            module_func_nodes[module_name][func_name].add(node.id)
            if len(node._labels) > 0:
                label = list(node._labels)[0]
            else:
                label = props.get("op_type", "unknown")
            name = node._properties.get("name", "?")
            G.add_node(node.id, xlabel=label, label=name, properties=node._properties)

        rels = list(results.graph()._relationships.values())
        for rel in rels:
            props = rel._properties
            label = rel.type
            G.add_edge(
                rel.start_node.id,
                rel.end_node.id,
                key=rel.id,
                label=label,
                type=rel.type,
                properties=props,
            )
        # logger.debug("G", G, dir(G))
        # logger.debug("mfn", module_func_nodes)
        ret = []
        for module_name, func_nodes in module_func_nodes.items():
            for func_name, nodes in func_nodes.items():
                G_ = G.subgraph(nodes).copy()
                attrs = {
                    "kind": "cfg",
                    "by": "isaac_toolkit.frontend.memgraph.llvm_mir_cdfg",
                    "module_name": module_name,
                    "func_name": func_name,
                }
                artifact = GraphArtifact(f"{legalize_str(module_name)}/{func_name}/llvm_cfg", G_, attrs=attrs)
                # print("artifact", artifact, dir(artifact), artifact.flags)
                ret.append(artifact)
    finally:
        session.close()

    # logger.debug("ret", ret)
    return ret


def get_dfg_artifacts(driver, label: str = "default"):
    query = Query(
        f"""
    MATCH (n)-[r:DFG]->(c)
    WHERE n.session = "{label}"
    RETURN *
    """
    )

    session = driver.session()
    try:
        results = session.run(query)
        # logger.debug("results", results)

        G = nx.MultiDiGraph()

        nodes = list(results.graph()._nodes.values())
        # logger.debug("nodes", nodes, len(nodes))
        module_func_bb_nodes = {}
        for node in nodes:
            props = node._properties
            module_name = props["module_name"]
            func_name = props["func_name"]
            bb_name = props["basic_block"]
            if module_name not in module_func_bb_nodes:
                module_func_bb_nodes[module_name] = {}
            if func_name not in module_func_bb_nodes[module_name]:
                module_func_bb_nodes[module_name][func_name] = {}
            if bb_name not in module_func_bb_nodes[module_name][func_name]:
                module_func_bb_nodes[module_name][func_name][bb_name] = set()
            module_func_bb_nodes[module_name][func_name][bb_name].add(node.id)
            if len(node._labels) > 0:
                label = list(node._labels)[0]
            else:
                label = props.get("op_type", "unknown")
            name = node._properties.get("name", "?")
            G.add_node(node.id, xlabel=label, label=name, properties=node._properties)

        rels = list(results.graph()._relationships.values())
        for rel in rels:
            props = rel._properties
            label = rel.type
            G.add_edge(
                rel.start_node.id,
                rel.end_node.id,
                key=rel.id,
                label=label,
                type=rel.type,
                properties=props,
            )
        # logger.debug("G", G, dir(G))
        # logger.debug("mfbn", module_func_bb_nodes)
        ret = []
        for module_name, func_bb_nodes in module_func_bb_nodes.items():
            for func_name, bb_nodes in func_bb_nodes.items():
                for bb_name, nodes in bb_nodes.items():
                    G_ = G.subgraph(nodes).copy()
                    attrs = {
                        "kind": "dfg",
                        "by": "isaac_toolkit.frontend.memgraph.llvm_mir_cdfg",
                        "module_name": module_name,
                        "func_name": func_name,
                        "bb_name": bb_name,
                    }
                    artifact = GraphArtifact(
                        f"{legalize_str(module_name)}/{func_name}/{bb_name}/llvm_dfg",
                        G_,
                        attrs=attrs,
                    )
                    # logger.debug("artifact", artifact, dir(artifact), artifact.flags)
                    ret.append(artifact)
    finally:
        session.close()

    # logger.debug("ret", ret)
    return ret


def load_cdfg(sess: Session, label: str = "default", force: bool = False):
    logger.info("Loading LLVM CDFG...")
    memgraph_config = sess.config.memgraph
    hostname = memgraph_config.hostname
    port = memgraph_config.port
    user = memgraph_config.user
    password = memgraph_config.password
    # TODO: database?

    driver = GraphDatabase.driver(f"bolt://{hostname}:{port}", auth=(user, password))
    try:
        cfgs = get_cfg_artifacts(driver, label=label)
        logger.debug("cfgs", cfgs)
        for cfg in cfgs:
            sess.add_artifact(cfg, override=force)
        dfgs = get_dfg_artifacts(driver, label=label)
        logger.debug("dfgs", dfgs)
        for dfg in dfgs:
            sess.add_artifact(dfg, override=force)
    finally:
        driver.close()


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    set_log_level(console_level=args.log, file_level=args.log)
    load_cdfg(sess, label=args.label, force=args.force)
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--label", default="default", required=True)
    parser.add_argument("--force", "-f", action="store_true")
    # TODO: allow overriding memgraph config?
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



#
# Copyright (c) 2025 TUM Department of Electrical and Computer Engineering.
#
# This file is part of ISAAC Toolkit.
# See https://github.com/tum-ei-eda/isaac-toolkit.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List, Set, Tuple, Optional, Union

import bisect

import sys
import logging
import argparse
import posixpath
from pathlib import Path
from collections import defaultdict

import pandas as pd
import numpy as np

from isaac_toolkit.session import Session
from isaac_toolkit.session.artifact import ArtifactFlag, filter_artifacts, TableArtifact
from isaac_toolkit.arch.riscv import riscv_branch_instrs, riscv_return_instrs
from isaac_toolkit.utils.demangle import unmangle_helper


logging.basicConfig(level=logging.DEBUG)  # TODO
logger = logging.getLogger(__name__)


PC_FUNC_NAME_CACHE = {}


# TODO: reset?


def find_func_name(mapping: Dict[str, Tuple[int, int]], pc: int) -> str:
    # TODO: refactor to use existing tables
    """
    Given a program counter, find the function it belongs to
    """
    # t0 = time.time()
    found = PC_FUNC_NAME_CACHE.get(pc)
    if found is not None:
        return found
    for func, ranges in mapping.items():
        for range_ in ranges:
            if pc >= range_[0] and pc <= range_[1]:
                PC_FUNC_NAME_CACHE[pc] = func
                return func
    ret = hex(pc)
    PC_FUNC_NAME_CACHE[pc] = ret
    return ret


# def collect_bbs(trace_df, mapping):
#     first_pc = None
#     # TODO: make this generic!
#     func2bbs = defaultdict(list)  # TODO: only track func_set?
#     # bb_freq = defaultdict(int)
#     prev_pc = None
#     prev_size = None
#     # prev_instr = None
#     bbs = []
#     # bb_instrs = []
#     num_bb_instrs = 0
#     bb_size = 0
#     # t0 = time.time()
#     trace_pcs = set(trace_df["pc"].unique())
#     # td = time.time() - t0
#     # print("td", td)
#     unique_bbs = []
#     # unique_bb_hashes = []
#     unique_bb_map = {}
#     unique_bb_freq = defaultdict(int)
#     bb_trace_data = []
#     # t0 = time.time()
#     trace_df["instr"] = trace_df["instr"].str.strip()  # TODO: fix in frontend
#     # td = time.time() - t0
#     # print("td_", td)
#
#     # t0 = time.time()
#     branch_ret_instrs = set(riscv_branch_instrs + riscv_return_instrs)
#     trace_df["is_branch_return"] = trace_df["instr"].isin(branch_ret_instrs)
#     # td = time.time() - t0
#     # print("td__", td)
#
#     # t0 = time.time()
#     pcs = trace_df["pc"].to_numpy()
#     instrs = trace_df["instr"].to_numpy()
#     sizes = trace_df["size"].to_numpy()
#     is_branch_return_list = trace_df["is_branch_return"].to_numpy()
#     # td = time.time() - t0
#     # print("td___", td)
#     # for row in trace_df.itertuples():
#     # Numpy based iter for speed
#
#     # t0 = time.time()
#     steps = pcs[1:] - pcs[:-1]
#     expected_steps = sizes[:-1]
#     irregular_step = steps != expected_steps
#     mask = is_branch_return_list.copy()
#     mask[:-1] |= irregular_step  # propagate irregular step to the earlier instr
#     bb_end_indices = np.where(mask)[0]
#     bb_start_indices = np.empty_like(bb_end_indices)
#     bb_start_indices[0] = 0
#     bb_start_indices[1:] = bb_end_indices[:-1] + 1
#     bbs = pd.DataFrame(
#         {
#             "first_pc": pcs[bb_start_indices],
#             "last_pc": pcs[bb_end_indices],
#             "num_instrs": bb_end_indices - bb_start_indices + 1,
#             "size": np.add.reduceat(sizes, bb_start_indices),  # grouped sum
#             "end_instr": instrs[bb_end_indices],
#         }
#     )
#     # td = time.time() - t0
#     # print("td___exp", td)
#     # print("irregular_step", irregular_step, sum(irregular_step))
#     # print("mask", mask, sum(mask))
#     # print("bbs", bbs)
#
#     # t0 = time.time()
#     tda = 0
#     tdb = 0
#     tdc = 0
#     tdd = 0
#     tde = 0
#     for i in range(len(pcs)):
#         # t0a = time.time()
#         pc = pcs[i]
#         instr = instrs[i]
#         sz = 2 if sizes[i] == 2 else 4
#         is_compressed = sz == 2
#         is_branch_return = is_branch_return_list[i]
#         trace_idx = i
#
#         # trace_idx = row.Index
#         # pc = row.pc
#         # instr = row.instr
#         # is_compressed = row.size == 2
#         # sz = 2 if is_compressed else 4
#         # tda += (time.time() - t0a)
#
#         # t0b = time.time()
#         if prev_pc:
#             step = pc - prev_pc
#             if step in [2, 4] and step == prev_size:
#                 pass
#             else:
#                 if first_pc is None:
#                     pass
#                 else:
#                     logger.warning("Detected potential trap @ pc = 0x%x -> 0x%x", prev_pc, pc)
#                     if True:
#                         func = find_func_name(mapping, prev_pc)
#                         # bb = BasicBlock(
#                         #     first_pc=first_pc,
#                         #     last_pc=prev_pc,
#                         #     num_instrs=len(bb_instrs),
#                         #     size=bb_size,
#                         #     end_instr=instr,
#                         #     func=func,
#                         # )
#                         # first_pc: int,
#                         # last_pc: int,
#                         # num_instrs: int,
#                         # size: int,
#                         # end_instr: str,
#                         # func: str,
#                         # bb_info = (first_pc, prev_pc, len(bb_instrs), bb_size, instr, func)
#                         bb_info = (first_pc, prev_pc, num_bb_instrs, bb_size, instr, func)
#                         # bbs.append(bb_info)
#                         # bb_hash = hash((first_pc, prev_pc))
#                         bb_hash = (first_pc << 32) | prev_pc
#                         ### bb_hash = bb.__hash__()
#                         # unique_bb_idx = None
#                         # if bb_hash in unique_bb_hashes:
#                         #     unique_bb_idx = unique_bb_hashes.index(bb_hash)
#                         unique_bb_idx = unique_bb_map.get(bb_hash)
#                         if unique_bb_idx is None:
#                             unique_bb_idx = len(unique_bbs)
#                             unique_bbs.append(bb_info)
#                             unique_bb_map[bb_hash] = unique_bb_idx
#                             func2bbs[func].append(unique_bb_idx)
#                         unique_bb_call = unique_bb_freq[unique_bb_idx]
#                         unique_bb_freq[unique_bb_idx] += 1
#
#                         bb_trace_data_new = (unique_bb_idx, unique_bb_call, trace_idx)
#                         bb_trace_data.append(bb_trace_data_new)
#
#                         # bb_instrs = []
#                         num_bb_instrs = 0
#                         bb_size = 0
#                         # bbs.append(bb)
#                         # if bb.get_freq() == 1:
#                         #     func2bbs[bb.func].append(bb)
#                         # bb_freq[bb] += 1
#
#                         first_pc = pc
#
#         # tdb += (time.time() - t0b)
#
#         # t0c = time.time()
#         # At the first pc of a basic block
#         if first_pc is None:
#             first_pc = pc
#
#         # bb_instrs.append(instr)
#         num_bb_instrs += 1
#         bb_size += sz
#         # tdc += (time.time() - t0c)
#
#         # t0d = time.time()
#         # if instr in riscv_branch_instrs + riscv_return_instrs:
#         if is_branch_return:
#             func = find_func_name(mapping, pc)
#             # bb = BasicBlock(
#             #     first_pc=first_pc,
#             #     last_pc=pc,
#             #     num_instrs=len(bb_instrs),
#             #     size=bb_size,
#             #     end_instr=instr,
#             #     func=func,
#             # )
#             # bb_info = (first_pc, pc, len(bb_instrs), bb_size, instr, func)
#             bb_info = (first_pc, pc, num_bb_instrs, bb_size, instr, func)
#             # bbs.append(bb_info)
#             # bb_hash = hash((first_pc, pc))
#             bb_hash = (first_pc << 32) | pc
#             # bb_hash = bb.__hash__()
#             # unique_bb_idx = None
#             # if bb_hash in unique_bb_hashes:
#             #     unique_bb_idx = unique_bb_hashes.index(bb_hash)
#             unique_bb_idx = unique_bb_map.get(bb_hash)
#             if unique_bb_idx is None:
#                 unique_bb_idx = len(unique_bbs)
#                 unique_bbs.append(bb_info)
#                 unique_bb_map[bb_hash] = unique_bb_idx
#                 func2bbs[func].append(unique_bb_idx)
#             unique_bb_call = unique_bb_freq[unique_bb_idx]
#             unique_bb_freq[unique_bb_idx] += 1
#
#             bb_trace_data_new = (unique_bb_idx, unique_bb_call, trace_idx)
#             bb_trace_data.append(bb_trace_data_new)
#             # bb_instrs = []
#             num_bb_instrs = 0
#             bb_size = 0
#             # bbs.append(bb)
#             # if bb.get_freq() == 1:
#             #     func2bbs[bb.func].append(bb)
#             #     bb_freq[bb] += 1
#             first_pc = None
#         # tdd += (time.time() - t0d)
#
#         # t0e = time.time()
#         prev_pc = pc
#         # prev_instr = instr
#         prev_size = sz
#         # tde += (time.time() - t0e)
#     # td = time.time() - t0
#     # print("tdloop", td, td/len(pcs))
#     # print("tda", tda, tda/td)
#     # print("tdb", tdb, tdb/td)
#     # print("tdc", tda, tdc/td)
#     # print("tdd", tdd, tdd/td)
#     # print("tde", tde, tde/td)
#     # t0 = time.time()
#     bb_trace_df = pd.DataFrame(bb_trace_data, columns=["bb_idx", "bb_call", "trace_idx"])
#     bb_trace_df["bb_idx"] = pd.to_numeric(bb_trace_df["bb_idx"], downcast="unsigned").astype("category")
#     bb_trace_df["bb_call"] = pd.to_numeric(bb_trace_df["bb_call"], downcast="unsigned")
#     bb_trace_df["trace_idx"] = pd.to_numeric(bb_trace_df["trace_idx"], downcast="unsigned")
#     # print("bb_trace_df", bb_trace_df)
#     # input("1")
#     unique_bbs_df = pd.DataFrame(unique_bbs,
#         columns=["first_pc", "last_pc", "num_instrs", "size", "end_instr", "func"])
#     unique_bbs_df["freq"] = list(unique_bb_freq.values())  # assume ordered dict
#     # td = time.time() - t0
#     # print("tdfin", td, td/len(pcs))
#     # input("!!!")
#     if first_pc is not None:
#         func = None
#
#     return trace_pcs, func2bbs, unique_bbs_df, bb_trace_df


def collect_bbs_new(trace_df, mapping):
    # --- Extract columns as NumPy ---
    pcs = trace_df["pc"].to_numpy()
    instrs = trace_df["instr"].str.strip().to_numpy()
    sizes = np.where(trace_df["size"].to_numpy() == 2, 2, 4)

    # --- Fast branch/return detection ---
    branch_ret_set = set(riscv_branch_instrs + riscv_return_instrs)
    is_branch_return = trace_df["instr"].isin(branch_ret_set).to_numpy()

    # --- Detect irregular steps ---
    steps = pcs[1:] - pcs[:-1]
    expected_steps = sizes[:-1]
    irregular_step = steps != expected_steps

    # --- Combine into a mask marking BB ends ---
    mask = is_branch_return.copy()
    mask[:-1] |= irregular_step

    # --- Indices of BB ends ---
    bb_end_indices = np.where(mask)[0]
    if len(bb_end_indices) == 0:
        return set(pcs), {}, pd.DataFrame(), pd.DataFrame()

    # --- Indices of BB starts ---
    bb_start_indices = np.empty_like(bb_end_indices)
    bb_start_indices[0] = 0
    bb_start_indices[1:] = bb_end_indices[:-1] + 1

    # --- Build unique BBs ---
    first_pcs = pcs[bb_start_indices]
    last_pcs = pcs[bb_end_indices]
    num_instrs = bb_end_indices - bb_start_indices + 1
    bb_sizes = np.add.reduceat(sizes, bb_start_indices)
    end_instrs = instrs[bb_end_indices]

    # Map PCs to funcs once
    unique_pcs = np.unique(np.concatenate([first_pcs, last_pcs]))
    func_map = {pc: find_func_name(mapping, pc) for pc in unique_pcs}
    funcs = [func_map[pc] for pc in last_pcs]

    # Deduplicate BBs using a deterministic key
    unique_bbs = []
    # unique_bb_map = {}
    func2bbs = defaultdict(list)

    bb_hashes = (first_pcs.astype(np.int64) << 32) | last_pcs
    unique_hashes, unique_idxs, inverse_idx = np.unique(bb_hashes, return_index=True, return_inverse=True)
    bb_idx_ = inverse_idx
    bb_idx_arr = np.empty(len(first_pcs), dtype=np.int32)
    bb_call_arr = np.empty(len(first_pcs), dtype=np.int32)
    trace_idx_arr = np.array(bb_end_indices, dtype=np.int32)
    unique_bb_freq = defaultdict(int)

    for unique_bb_idx in range(len(unique_hashes)):
        idx = unique_idxs[unique_bb_idx]
        fp = first_pcs[idx]
        lp = last_pcs[idx]
        n = num_instrs[idx]
        sz = bb_sizes[idx]
        ei = end_instrs[idx]
        func = funcs[idx]
        unique_bbs.append((fp, lp, n, sz, ei, func))
        func2bbs[func].append(unique_bb_idx)

    for idx in range(len(first_pcs)):
        # unique_bb_idx = bb_hash
        unique_bb_idx = bb_idx_[idx]

        bb_call = unique_bb_freq[unique_bb_idx]
        unique_bb_freq[unique_bb_idx] += 1
        bb_idx_arr[idx] = unique_bb_idx
        bb_call_arr[idx] = bb_call
        trace_idx_arr[idx] = bb_end_indices[idx]

    # --- Convert to DataFrames ---
    bb_trace_df = pd.DataFrame(
        {
            "bb_idx": bb_idx_arr,  # .astype("category"),
            "bb_call": bb_call_arr.astype(np.uint32),
            "trace_idx": trace_idx_arr.astype(np.uint32),
        }
    )
    bb_trace_df["bb_idx"] = bb_trace_df["bb_idx"].astype("category")
    bb_trace_df["bb_call"] = pd.to_numeric(bb_trace_df["bb_call"], downcast="unsigned")
    bb_trace_df["trace_idx"] = pd.to_numeric(bb_trace_df["trace_idx"], downcast="unsigned")

    unique_bbs_df = pd.DataFrame(unique_bbs, columns=["first_pc", "last_pc", "num_instrs", "size", "end_instr", "func"])
    unique_bbs_df["freq"] = [unique_bb_freq[i] for i in range(len(unique_bbs))]

    return set(pcs), func2bbs, unique_bbs_df, bb_trace_df


# TODO: consistent arg naming
# TODO: re-format quotes


# def callgrind_format_get_inclusive_cost(bbs: List[BasicBlock]):
def callgrind_format_get_inclusive_cost(bb_trace_df: pd.DataFrame, unique_bbs_df: pd.DataFrame):
    call_stack = []
    bb_stack = []
    total_cost = 0
    inclusive_cost_dict = defaultdict(lambda: defaultdict(list))
    prev_bb_idx = None
    # unique_bbs_records = unique_bbs_df.to_records(index=False)

    # call_stack: [A, B]
    # [Given] bb_stack: [[bb1, bb2, bb3], [bb4, bb5, bb6]]
    # [B returns]: [[bb1, bb2, [bb3, # cost from B]]]
    # where bb1 - bb3 belong to func A and bb4 - bb6 belong to func B
    # for i, bb in enumerate(bbs):
    # for i, bb_trace_row in bb_trace_df.iterrows():
    bb_funcs = unique_bbs_df["func"].values  # dtype=object or better: categorical/int
    bb_num_instrs = unique_bbs_df["num_instrs"].values  # dtype=int32/int64
    bb_end_instrs = unique_bbs_df["end_instr"].values  # dtype=object or categorical
    bb_first_pc = unique_bbs_df["first_pc"].values
    bb_last_pc = unique_bbs_df["last_pc"].values
    # t0 = time.time()
    # ts = [t0]
    # for i, bb_idx in enumerate(bb_trace_df["bb_idx"].values):
    for _, bb_idx in enumerate(bb_trace_df["bb_idx"].values):
        # ts.append(time.time())
        # if i == 100:
        #     break
        func = bb_funcs[bb_idx]
        n_instr = bb_num_instrs[bb_idx]
        # end_instr = bb_end_instrs[bb_idx]
        # first_pc = bb_first_pc[bb_idx]
        # last_pc = bb_last_pc[bb_idx]
        # print("progress", float(i)/len(bb_trace_df))
        # bb_idx = bb_trace_row.bb_idx
        # bb = unique_bbs_df.iloc[bb_idx]
        # bb = unique_bbs_records[bb_idx]
        # print("i,bb_idx,bb", i, bb_idx, bb)
        total_cost += n_instr
        # print("bb", bb)
        # print("prev_bb", prev_bb)
        if prev_bb_idx is None or (bb_end_instrs[prev_bb_idx] in riscv_branch_instrs and bb_funcs[prev_bb_idx] != func):
            # first bb in the trace
            # print("append call_stack + bb_stack")
            call_stack.append(func)
            bb_stack.append([bb_idx])
        elif bb_funcs[prev_bb_idx] == func:
            # jalr doesn't necessarily mean return
            # 0x2ac8, jalr, memset -> 0x2ae8, memset
            # print("append bb_stack")
            bb_stack[-1].append(bb_idx)
        # elif prev_bb.end_instr in return_instrs:
        # elif prev_bb.end_instr in riscv_return_instrs:
        elif bb_end_instrs[prev_bb_idx] in riscv_return_instrs or bb_num_instrs[prev_bb_idx].num_instrs == 1:
            # Check whether jalr refer to return
            # sometimes jalr simply means indirect jump
            # TODO: Redundant? Is it already handled in the above condition?
            if func not in call_stack:
                call_stack.append(func)
                bb_stack.append([bb_idx])
                # print("append call_stack + bb_stack 2")
                continue

            # ret case for the rest code snippet
            diff = len(call_stack) - call_stack.index(func) - 1

            for j in range(diff):
                cost = 0
                callee_first_bb_idx = bb_stack[-1][0][0] if isinstance(bb_stack[-1][0], list) else bb_stack[-1][0]
                for bb_stack_elem in bb_stack[-1]:
                    if not isinstance(bb_stack_elem, list):
                        # cost += 1 + (bb_stack_elem.last_pc - bb_stack_elem.first_pc) // 4
                        cost += bb_num_instrs[bb_stack_elem]
                    else:
                        subroutine_cost = bb_stack_elem[1]
                        cost += (
                            # subroutine_cost + 1 + (bb_stack_elem[0].last_pc - bb_stack_elem[0].first_pc) // 4
                            subroutine_cost
                            + bb_num_instrs[bb_stack_elem[0]]
                        )

                # print("pop call_stack + bb_stack")
                bb_stack.pop()
                call_stack.pop()

                caller = bb_stack[-1][-1]
                if isinstance(caller, list):
                    caller_bb_idx = caller[0]
                    inclusive_cost_dict[bb_last_pc[caller_bb_idx]][bb_first_pc[callee_first_bb_idx]].append(cost)
                    bb_stack[-1][-1][1] += cost
                else:
                    caller_bb_idx = caller
                    inclusive_cost_dict[bb_last_pc[caller_bb_idx]][bb_first_pc[callee_first_bb_idx]].append(cost)
                    bb_stack[-1][-1] = [caller_bb_idx, cost]

                # When callee function returns, it jumps to the start of the "next" basic block
                # in caller function. For example:
                # 0x6b0  jal    rd, imm <--- in function A
                # ...
                # 0x5ae0 jalr   rd, rs, imm <--- in function B
                # 0x6b4  ... <--- return to function A
                if j == diff - 1:
                    # print("append bb stack")
                    bb_stack[-1].append(bb_idx)

        prev_bb_idx = bb_idx
    # print("ts", ts)
    # import numpy as np
    # tds = np.array(ts)[1:] - np.array(ts)[:-1]
    # print("tds", tds)
    # input()

    # print("inclusive_cost_dict", inclusive_cost_dict)
    # inclusive_cost_dict_sum = {pc: sum([sum(x) for x in costs.values()]) for pc, costs in inclusive_cost_dict.items()}
    # print("inclusive_cost_dict_sum", inclusive_cost_dict_sum)
    # inclusive_cost_dict_total = sum(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_total", inclusive_cost_dict_total)
    # inclusive_cost_dict_max = max(inclusive_cost_dict_sum.values())
    # print("inclusive_cost_dict_sum_max", inclusive_cost_dict_max)
    # print("total_cost", total_cost)
    # TODO: check that all costs are contained

    return inclusive_cost_dict


def callgrind_format_converter(
    # bbs: List[BasicBlock],
    bb_trace_df: pd.DataFrame,
    unique_bbs_df: pd.DataFrame,
    trace_pcs: Set[int],
    mapping: Dict[str, Tuple[int, int]],
    # func2bbs: Dict[str, List[BasicBlock]],
    func2bbs: Dict[str, List[int]],
    # bb_freq: Dict[BasicBlock, int],
    srcFile_func_dict: Dict[str, List[str]],
    srcFile_linkage_name_dict: Dict[str, List[str]],
    srcFile_unmangled_linkage_name_dict: Dict[str, List[str]],
    func_set: Set[str],
    file2pc_loc: Optional[Dict[str, List[Tuple[int, int]]]] = None,
    dump_pc: bool = True,
    dump_pos: bool = False,
    elf_file_path: str = "/path/to/elf",
    unmangle_names: bool = False,
):
    """
    Collect necessary information for callgrind output format.
    Then dump it to output_file

    Necessary information is:
    1. function name and its source file name
    Output:
        ...
        fl={source_file_name}
        fn={func_name}

    2. If --dump-instr=yes
    Output:
        ...
        {pc} {frequency}
        ...

    3. If --dump-line=yes
    Output:
        ...
        {line_number_in_source_code} {frequency}
        ...

    4. Caller-callee relationship
    Output:
        ... in caller's context
        {pc} {frequency}
        cfi={callee's source file}
        calls={number of times caller calls callee} {callee pc}
        {pc} {inclusive_cost}
    """
    assert dump_pc or dump_pos
    # inclusive_cost_dict = callgrind_format_get_inclusive_cost(bbs)
    inclusive_cost_dict = callgrind_format_get_inclusive_cost(bb_trace_df, unique_bbs_df)

    # Find the source file where the function is defined
    # Some functions come from libc. In this case ??? is returned.
    def find_srcFile(target_func: str) -> str:
        for srcFile, funcs in srcFile_func_dict.items():
            for func in funcs:
                if func == target_func:
                    return srcFile
        for srcFile, funcs in srcFile_linkage_name_dict.items():
            for func in funcs:
                if func == target_func:
                    return srcFile
        return "???"

    #
    def aggregate_pos_cost_of_func(sorted_bb_lists: list):
        position_cost_dict = defaultdict(int)
        for bb_idx in sorted_bb_lists:
            bb = unique_bbs_df.iloc[bb_idx]
            pc = bb.first_pc
            # while pc <= bb.last_pc:
            for pc_ in range(pc, bb.last_pc + 2, 2):
                if pc_ not in trace_pcs:
                    # pc += 2
                    continue
                # position_cost_dict[pc_] += bb.get_freq()
                position_cost_dict[pc_] += bb.freq
                # pc += 4
                # pc += 2
        return position_cost_dict

    positions = "instr" if dump_pc else "line"

    prologue = f"""\
# callgrind format
version: 1
creator: callgrind-3.15.0
pid: 2577934
cmd:  ./dhrystone
part: 1

desc: I1 cache:
desc: D1 cache:
desc: LL cache:

desc: Timerange:
desc: Trigger: Program termination

positions: {positions}
events: Ir
summary:

"""
    callgrind_output = ""

    def callgrind_format_dump_instr(pc: int, source_file: str) -> str:
        return hex(pc)

    def callgrind_format_dump_line(pc: int, source_file: str) -> str:
        if source_file == "???":
            return "0"
        assert file2pc_loc is not None
        mapping_ = file2pc_loc.get(source_file)
        if mapping_ is None:
            return "0"
        # PYTHON 3.10:
        # i = bisect.bisect_right(mapping_, pc, key=lambda x: x[0])
        # if i:
        #     start_pc, line = mapping_[i - 1]
        #     if pc >= start_pc:
        #         return f"{line}"
        # ---
        # PYTHON 3.8
        mapping_first = [x[0] for x in mapping_]
        i = bisect.bisect_right(mapping_first, pc)
        if i:
            start_pc, line = mapping_[i - 1]
            if pc >= start_pc:
                return f"{line}"
        # ---
        return "0"

    dump_positions = callgrind_format_dump_instr if dump_pc else callgrind_format_dump_line

    # source file to functions mapping
    srcFile_to_func = defaultdict(list)
    for func in func_set:
        srcFile_to_func[find_srcFile(func)].append(func)

    for srcFile, funcs in srcFile_to_func.items():
        callgrind_output += f"ob={posixpath.abspath(elf_file_path)}\n"
        callgrind_output += f"fl={srcFile}\n"

        for func in funcs:

            bb_list = func2bbs[func]
            bb_list.sort(key=lambda bb_idx: unique_bbs_df.iloc[bb_idx].first_pc)
            if unmangle_names:
                func = unmangle_helper(func)
            callgrind_output += f"fn={func}\n"

            position_cost_dict = aggregate_pos_cost_of_func(bb_list)

            # branch_pc_list = [bb.last_pc for bb in bb_list]
            branch_pc_list = unique_bbs_df.iloc[bb_list].last_pc.values

            for pc in sorted(position_cost_dict.keys()):
                position_info = dump_positions(pc, srcFile)
                callgrind_output += f"{position_info} {position_cost_dict[pc]}\n"
                if pc in branch_pc_list and pc in inclusive_cost_dict:
                    for callee_pc, inclusive_cost in inclusive_cost_dict[pc].items():
                        # TODO Share object files case not implemented here
                        callgrind_output += f"cob={posixpath.abspath(elf_file_path)}\n"
                        callee_func = find_func_name(mapping, callee_pc)
                        callgrind_output += f"cfi={find_srcFile(callee_func)}\n"
                        if unmangle_names:
                            callee_func = unmangle_helper(callee_func)
                        callgrind_output += f"cfn={callee_func}\n"
                        callgrind_output += f"calls={len(inclusive_cost)} {hex(callee_pc)}\n"
                        callgrind_output += f"{position_info} {sum(inclusive_cost)}\n"

            callgrind_output += "\n"

    content = prologue + callgrind_output
    return content


def collect_trace_bbs(
    sess: Session,
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
):
    artifacts = sess.artifacts
    # elf_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.ELF)
    # assert len(elf_artifacts) == 1
    # elf_artifact = elf_artifacts[0]

    trace_artifacts = filter_artifacts(artifacts, lambda x: x.flags & ArtifactFlag.INSTR_TRACE)
    assert len(trace_artifacts) == 1
    trace_artifact = trace_artifacts[0]

    func2pc_artifacts = filter_artifacts(artifacts, lambda x: x.name == "func2pc")
    assert len(func2pc_artifacts) == 1
    func2pc_artifact = func2pc_artifacts[0]
    func2pc_df = func2pc_artifact.df

    mapping = func2pc_df.groupby("func")["pc_range"].apply(list).to_dict()
    # bbs, trace_pcs, func2bbs, bb_freq = collect_bbs(trace_artifact.df, mapping)
    # t0 = time.time()
    # trace_pcs, func2bbs, unique_bbs_df, bb_trace_df = collect_bbs(trace_artifact.df, mapping)
    # t1 = time.time()
    trace_pcs, func2bbs, unique_bbs_df, bb_trace_df = collect_bbs_new(trace_artifact.df, mapping)
    # t2 = time.time()
    # tdx = t1 - t0
    # tdy = t2 - t1
    # print("tdx", tdx)
    # print("tdy", tdy)
    # print("bb_trace_df", bb_trace_df)
    # print("bb_trace_df2", bb_trace_df2)
    # print("unique_bbs_df", unique_bbs_df)
    # print("unique_bbs_df2", unique_bbs_df2)

    bb_trace_attrs = {
        "trace": trace_artifact.name,
        "kind": "bb_trace",
        "by": __name__,
    }

    unique_bbs_attrs = {
        "trace": trace_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    func2bbs_attrs = {
        "trace": trace_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    trace_pcs_attrs = {
        "trace": trace_artifact.name,
        "kind": "list",
        "by": __name__,
    }

    trace_pcs_df = pd.Series(list(trace_pcs))  # TODO: to_frame()?
    # print("trace_pcs_df", trace_pcs_df)

    # print("func2bbs", func2bbs)
    func2bbs_df = pd.DataFrame({"func": list(func2bbs.keys()), "bb_idxs": list(func2bbs.values())})
    # print("func2bbs_df", func2bbs_df)

    bb_trace_artifact = TableArtifact("bb_trace", bb_trace_df, attrs=bb_trace_attrs)
    unique_bbs_artifact = TableArtifact("unique_bbs", unique_bbs_df, attrs=unique_bbs_attrs)
    func2bbs_artifact = TableArtifact("func2bbs", func2bbs_df, attrs=func2bbs_attrs)
    trace_pcs_artifact = TableArtifact("trace_pcs", trace_pcs_df, attrs=trace_pcs_attrs)

    sess.add_artifact(bb_trace_artifact, override=force)
    sess.add_artifact(unique_bbs_artifact, override=force)
    sess.add_artifact(func2bbs_artifact, override=force)
    sess.add_artifact(trace_pcs_artifact, override=force)

    # from basic_blocks (for backward compat)
    pc2bb_df = unique_bbs_df.copy()
    # TODO: check if inclusive or exclusive range
    # pc2bb_df.rename(columns={"func": "func_name", "first_pc": "start", "last_pc": "end"}, inplace=True)
    pc2bb_df.rename(columns={"first_pc": "start", "last_pc": "end"}, inplace=True)
    pc2bb_df["func_name"] = pc2bb_df["func"].map(lambda x: {x})
    pc2bb_df.sort_values("freq", inplace=True, ascending=False)
    pc2bb_df["weight"] = pc2bb_df["freq"] * pc2bb_df["num_instrs"]
    pc2bb_df["rel_weight"] = pc2bb_df["weight"] / sum(pc2bb_df["weight"])
    # TODO: size is wrong!!!
    pc2bb_df.drop(columns=["end_instr", "func"], inplace=True)

    attrs = {
        "trace": trace_artifact.name,
        "kind": "mapping",
        "by": __name__,
    }

    # pc2bb_artifact = TableArtifact("pc2bb_compat", pc2bb_df, attrs=attrs)
    pc2bb_artifact = TableArtifact("pc2bb", pc2bb_df, attrs=attrs)
    sess.add_artifact(pc2bb_artifact, override=force)


def handle(args):
    assert args.session is not None
    session_dir = Path(args.session)
    assert session_dir.is_dir(), f"Session dir does not exist: {session_dir}"
    sess = Session.from_dir(session_dir)
    collect_trace_bbs(
        sess,
        force=args.force,
    )
    sess.save()


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    parser.add_argument("--session", "--sess", "-s", type=str, required=True)
    parser.add_argument("--force", "-f", action="store_true")
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import sys
import argparse
import subprocess
from typing import Union
from pathlib import Path

import yaml


def dot2pdf_helper(in_file: Union[str, Path], out_file: Union[str, Path]):
    with open(out_file, "wb") as f:
        dot_args = ["dot", "-Tpdf", in_file]
        print(">", " ".join(map(str, dot_args)))
        _ = subprocess.run(dot_args, check=True, stdout=f)
    print(f"Converted {in_file} -> {out_file}")


def combine_pdfs(index, output=None):
    with open(index, "r") as f:
        combined_index_data = yaml.safe_load(f)

    pdf_files = [
        Path(candidate_data["artifacts"]["io_sub"].replace(".pkl", ".pdf"))
        for candidate_data in combined_index_data["candidates"]
    ]
    dot_files = [
        Path(candidate_data["artifacts"]["io_sub"].replace(".pkl", ".dot"))
        for candidate_data in combined_index_data["candidates"]
    ]
    for i, pdf_file in enumerate(pdf_files):
        if not pdf_file.is_file():
            dot_file = dot_files[i]
            assert dot_file.is_file()
            dot2pdf_helper(dot_file, pdf_file)
    assert len(pdf_files) > 0, "No files found!"
    assert output is not None
    out_file = Path(output)
    combine_args = ["pdfunite", *pdf_files, out_file]
    # print(">", " ".join(map(str, combine_args)))
    _ = subprocess.run(combine_args, capture_output=True, text=True, check=True)
    # print(f"Wrote combined PDF file: {out_file}")


def handle(args):
    combine_pdfs(args.index, output=args.output)


def get_parser():
    parser = argparse.ArgumentParser(description="Extract dot files from index and merge graphs into single pdf")
    parser.add_argument("index", help="Index yaml file")
    parser.add_argument("-o", "--output", default=None, help="Output yaml file")
    parser.add_argument(
        "--log",
        default="info",
        choices=["critical", "error", "warning", "info", "debug"],
    )  # TODO: move to defaults
    return parser


def main(argv):
    parser = get_parser()
    args = parser.parse_args(argv)
    handle(args)


if __name__ == "__main__":
    main(sys.argv[1:])
